# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: projection_shift
   :platform: Unix
   :synopsis: Calculate horizontal and vertical shifts in the projection\
       images over time, using template matching.

.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>

"""

import logging
import numpy as np
from skimage.feature import match_template, match_descriptors, ORB
from scipy.linalg import lstsq
from skimage.transform import AffineTransform
from skimage.measure import ransac

from savu.plugins.utils import register_plugin
from savu.plugins.filters.base_filter import BaseFilter
from savu.plugins.driver.cpu_plugin import CpuPlugin


@register_plugin
class ProjectionShift(BaseFilter, CpuPlugin):
    """
    Horizontal and vertical shift are calculated using a chosen method and \
    added to the metadata.  The vertical and horizontal shifts can be \
    corrected using the ProjectionVerticalAlignment and SinogramAlignment \
    (in 'shift' mode) plugins respectively.

    Method: Uses either skimage template_matching or orb feature tracking plus\
    robust ransac matching to calculate the translation between different\
    combinations of 10 consecutive projection images. A least squares solution\
    to the shift values between images is calculated and returned for the\
    middle 8 images.

    :param method: Method used to calculate the shift between images. Choose \
        from 'template_matching' and 'orb_ransac'. Default: 'orb_ransac'.
    :param template: Position of the template to match (required) \
        e.g. [300:500, 300:500]. Default: None.
    :param threshold: e.g. [a, b] will set all values above a to \
        b. Default: None.
    :param n_keypoints: Number of keypoints to use in ORB feature \
        detector.  Default: 20.
    :param out_datasets: Set the output dataset name. Default: ['proj_shift'].

    :config_warn: The template parameter is required and must not be None.
    """

    def __init__(self):
        logging.debug("initialising Sinogram Alignment")
        super(ProjectionShift, self).__init__("ProjectionShift")
        self.template = None
        self.threshold = 0

    def pre_process(self):
        if self.parameters['method'] == 'template_matching':
            self.template_params = []
            for p in self.parameters['template']:
                start, end = p.split(':')
                self.template_params.append(slice(int(start), int(end)))
            self._calculate_shift = self._template_matching_shift
        elif self.parameters['method'] == 'orb_ransac':
            self._calculate_shift = self._orb_ransac_shift

        if self.parameters['threshold']:
            self.threshold = self.parameters['threshold']

        self.sl = [slice(None)]*3
        self.sl2 = [slice(None)]*3
        self.slice_dir = self.get_plugin_in_datasets()[0].get_slice_dimension()
        self.A = self._calculate_frame_matrix()

    def _calculate_frame_matrix(self):
        n_unknowns = self.get_max_frames() + 2  # 2 padded frames
        frame_list = self._calculate_frame_list(np.arange(n_unknowns))
        n_equations = len(frame_list)
        A = np.zeros((n_equations, n_unknowns))
        for i in range(len(frame_list)):
            for f in frame_list[i][1:]:
                A[i, f] = 1
        return A

    def process_frames(self, data):
        data, nFrames, output, shift_array = self._initial_setup(data)
        return self._sub_pixel_shift_adjustment(data)

    def _initial_setup(self, data):
        data = data[0]
        shape = list(data.shape)
        nFrames = data.shape[self.slice_dir]-2
        shape[self.slice_dir] += -2
        output = np.zeros(tuple(shape))
        shift_array = np.zeros((nFrames, 2))
        return data, nFrames, output, shift_array

    def _get_shift(self, data, frame1, frame2):
        self.sl[self.slice_dir] = frame1
        self.sl2[self.slice_dir] = frame2

        d1 = data[self.sl]
        d2 = data[self.sl2]
        if self.template:
            self.template = data[self.sl][self.template_params]

        if self.threshold:
            d1[d1 > self.threshold[0]] = self.threshold[1]
            d2[d2 > self.threshold[0]] = self.threshold[1]
            if self.template:
                self.template[self.template > self.threshold[0]] = \
                    self.threshold[1]
        return self._calculate_shift(d1, d2, self.template)

    def _orb_ransac_shift(self, im1, im2, template):
        descriptor_extractor = ORB() #n_keypoints=self.parameters['n_keypoints'])
        key1, des1 = self._find_key_points(descriptor_extractor, im1)
        key2, des2 = self._find_key_points(descriptor_extractor, im2)
        matches = match_descriptors(des1, des2, cross_check=True)

        # estimate affine transform model using all coordinates
        src = key1[matches[:, 0]]
        dst = key2[matches[:, 1]]

        # robustly estimate affine transform model with RANSAC
        model_robust, inliers = ransac((src, dst), AffineTransform,
                                       min_samples=3, residual_threshold=1,
                                       max_trials=100)
#        diff = []
#        for p1, p2 in zip(src[inliers], dst[inliers]):
#            diff.append(p2-p1)
#        return np.mean(diff, axis=0)

        return model_robust.translation

    def _find_key_points(self, desc_extractor, image):
        desc_extractor.detect_and_extract(image)
        keypoints = desc_extractor.keypoints
        descriptors = desc_extractor.descriptors
        return keypoints, descriptors

    def _template_matching_shift(self, im1, im2, template):
        index = []
        for im in [im1, im2]:
            match = match_template(im, template)
            index.append(np.unravel_index(np.argmax(match), match.shape))
        index = np.array(index)
        shift = index[1] - index[0]
        return shift

    def _sub_pixel_shift_adjustment(self, data):
        frame_list = \
            self._calculate_frame_list(np.arange(data.shape[self.slice_dir]))

        new_shift = []
        for f in frame_list:
            new_shift.append(
                self._get_shift(data, f[0], f[-1]).astype(np.float64))

        return self._calculate_new_shift_array(np.array(new_shift))

    def _calculate_frame_list(self, frames):
        sixes = list(zip(*(frames[i:] for i in range(6))))
        fives = list(zip(*(frames[i:] for i in range(5))))
        fours = list(zip(*(frames[i:] for i in range(4))))
        threes = list(zip(*(frames[i:] for i in range(3))))
        return sixes + fives + fours + threes

    def _calculate_new_shift_array(self, shift):
        new_shift = []
        for i in range(2):
            new_shift.append(lstsq(self.A, shift[:, i])[0])
        return np.transpose(np.array(new_shift))[1:-1]

    def post_process(self):
        out_data = self.get_out_datasets()[0]
        self.get_in_datasets()[0].meta_data.set(
            'proj_align_shift_local', out_data.data[:, :])
        self.get_in_datasets()[0].meta_data.set(
            'proj_align_shift', np.cumsum(out_data.data[:, :], axis=0))

    def get_max_frames(self):
        # Do not change this number as 8 is currently a requirement.
        return 8

    def nOutput_datasets(self):
        return 1

    def setup(self):
        # set up the output dataset that is created by the plugin
        in_dataset, out_dataset = self.get_datasets()

        in_pData, out_pData = self.get_plugin_datasets()
        in_pData[0].plugin_data_setup('PROJECTION', self.get_max_frames(),
                                      fixed=True)

        new_shape = (in_dataset[0].get_shape()[
            in_dataset[0].get_slice_directions()[0]], 2)

        out_dataset[0].create_dataset(shape=new_shape,
                                      axis_labels=['x.pixels', 'y.pixels'],
                                      remove=True)
        out_dataset[0].add_pattern("METADATA", core_dims=(1,), slice_dims=(0,))
        out_pData[0].plugin_data_setup('METADATA', self.get_max_frames(),
                                       fixed=True)

    def set_filter_padding(self, in_data, out_data):
        pad_dim = in_data[0].get_slice_directions()[0]
        in_data[0].padding = {'pad_directions': [str(pad_dim) + '.1']}
        #in_data[0].padding = {'pad_directions': [str(pad_dim) + '.before.1']}



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: nxstxm_loader
   :platform: Unix
   :synopsis: A class for loading nxstxm data
.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>
"""

from savu.plugins.loaders.mapping_loaders.base_multi_modal_loader \
    import BaseMultiModalLoader
import numpy as np
from savu.plugins.utils import register_plugin
import logging
import h5py


@register_plugin
class I13PtychoLoader(BaseMultiModalLoader):
    """
    A class to load tomography data from an NXstxm file
    :param mono_energy: The mono energy. Default: 9.1.
    :param is_tomo: The mono energy. Default: True.
    :param theta_step: The theta step. Default:1.0.
    :param theta_start: The theta start. Default: -90.0.
    :param theta_end: The theta end. Default: 90.0.
    """

    def __init__(self, name='I13PtychoLoader'):
        super(I13PtychoLoader, self).__init__(name)

    def setup(self):
        """
         Define the input nexus file
        :param path: The full path of the NeXus file to load.
        :type path: str
        """

        exp = self.exp
        data_obj = exp.create_data_object("in_data", 'ptycho')
        data_obj.backing_file = \
            h5py.File(exp.meta_data.get("data_file"), 'r')
        logging.debug("Creating file '%s' '%s'_entry",
                      data_obj.backing_file.filename, 'ptycho')

        data_obj.data = data_obj.backing_file['entry1/instrument/merlin_sw_hdf/data']
#        print data_obj.data.shape
        data_obj.set_shape(data_obj.data.shape)
        try:
            control = data_obj.backing_file['/entry1/instrument/ionc_i/ionc_i'][()]
        # this is global since it is to do with the beam
            exp.meta_data.set("control", control)
            logging.debug('adding the ion chamber to the meta data')
        except:
            logging.warning('No ion chamber information. Leaving this blank')

        self.exp.meta_data.set("mono_energy", self.parameters['mono_energy'])

        labels = []
        ### set the rotation
        rotation_angle = None
        if self.parameters['is_tomo']:
            rotation_angle = np.arange(self.parameters['theta_start'],self.parameters['theta_end'], self.parameters['theta_step'])
            labels.append('rotation_angle.degrees')
            data_obj.meta_data.set('rotation_angle', rotation_angle)

#             try:
#                 rotation_angle = data_obj.backing_file['entry1/instrument/t1_theta/t1_theta'][()]
#                 if rotation_angle.ndim > 1:
#                     rotation_angle = rotation_angle[:, 0]
#                 # axis label
#                 labels.append('rotation_angle.degrees')
#
#                 data_obj.meta_data.set('rotation_angle', rotation_angle)
#             except KeyError:
#                 logging.debug("Not a tomography!")

        ### GET THE AXES ###
        x = data_obj.backing_file['entry1/instrument/lab_sxy/lab_sx'][()]*1e-6
        data_obj.meta_data.set('x', x)
        y = data_obj.backing_file['entry1/instrument/lab_sxy/lab_sy'][()]*1e-6
        data_obj.meta_data.set('y', y)
        if rotation_angle is not None and len(x.shape)>1:
            pos = np.zeros((x.shape[0],2,x.shape[1]))
            pos[:,0,:] = y
            pos[:,1,:] = x
        else:
            pos = np.zeros((2,len(y)))
            pos[0,:] = y
            pos[1,:] = x
        data_obj.meta_data.set('xy', pos)
        ######

        ### NOW DO THE LABELS
        labels.append('xy.metres')

        labels.extend(['detectorX.pixel','detectorY.pixel'])
        logging.debug('The labels are: %s',labels)

        data_obj.set_axis_labels(*tuple(labels))
        dims = list(range(len(data_obj.get_shape())))
        diff_core = (-2,-1) # it will always be this
        diff_slice = tuple(dims[:-2])
        logging.debug("is a diffraction")
        logging.debug("the diffraction cores are:"+str(diff_core))
        logging.debug("the diffraction slices are:"+str(diff_slice))
        positions_label = (data_obj.get_data_dimension_by_axis_label('xy', contains=True),)
        rotation_label=[]
        if self.parameters['is_tomo']:
            rotation_label = (data_obj.get_data_dimension_by_axis_label('rotation_angle', contains=True),)
            sino_cores = rotation_label + positions_label
            data_obj.add_pattern("SINOGRAM", core_dims=sino_cores, slice_dims=tuple(set(dims)-set(sino_cores)))

        data_obj.add_pattern("PROJECTION", core_dims=positions_label, slice_dims=tuple(set(dims)-set(positions_label)))



        data_obj.add_pattern("DIFFRACTION", core_dims=diff_core,
                             slice_dims=diff_slice)

        data_obj.add_pattern("4D_SCAN", core_dims=tuple(set(dims)-set(rotation_label)),
                        slice_dims=rotation_label)

        self.set_data_reduction_params(data_obj)



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: basic_operations
   :platform: Unix
   :synopsis: Plugin to perform basic mathematical operations on datasets.

.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>

"""

from savu.plugins.plugin import Plugin
from savu.plugins.utils import register_plugin
from savu.plugins.driver.cpu_plugin import CpuPlugin


@register_plugin
class BasicOperations(Plugin, CpuPlugin):
    """ A class that performs basic mathematical operations on datasets.
    How should the information be passed to the plugin?

    :param operations: Operations to perform. Default: [].
    :param pattern: Pattern associated with the \
        datasets. Default: 'PROJECTION'.
    """

    def __init__(self):
        super(BasicOperations, self).__init__("BasicOperations")

    def pre_process(self):
        self.operations = self._amend_ops(self._set_data_mappings())
        self.out_data = self._set_out_data_names()

    def process_frames(self, data):
        # creates an 'environment' that will store the variables created inside the exec statement
        exec_environment = {'data': data}

        for i in range(len(self.operations)):
            # runs the exec with no builtins, and only 'data' available as a variable initially
            exec(f"{self.out_data[i]} = {self.operations[i]}", {"builtins": None}, exec_environment)

        # Find the result from each exec. Does list comprehension on the results instead of just
        # exec_environment.items to keep the order the same as in out_data
        return [exec_environment[out] for out in self.out_data]

    def setup(self):
        """
        Initial setup of all datasets required as input and output to the \
        plugin.  This method is called before the process method in the \
        plugin chain.
        """

        in_datasets, out_datasets = self.get_datasets()
        in_pData, out_pData = self.get_plugin_datasets()
        pattern = self.parameters['pattern']

        for pData in in_pData:
            pData.plugin_data_setup(pattern, self.get_max_frames())

        # making the assumption that the basic operations do not change the
        # shape of the data for now.
        copy_datasets = self._get_associated_datasets()
        for i in range(len(out_datasets)):
            out_datasets[i].create_dataset(in_datasets[copy_datasets[i]])
            out_pData[i].plugin_data_setup(pattern, self.get_max_frames())

    def nInput_datasets(self):
        return 'var'

    def nOutput_datasets(self):
        return 'var'

    def get_max_frames(self):
        return 'multiple'

    def _set_data_mappings(self):
        mapping_dict = {}
        in_datasets = self.get_in_datasets()
        for i in range(len(in_datasets)):
            mapping_dict[in_datasets[i].get_name()] = 'data[' + str(i) + ']'
        return mapping_dict

    def _set_out_data_names(self):
        out_datasets = self.get_out_datasets()
        return [out_datasets[i].get_name() for i in range(len(out_datasets))]

    def _amend_ops(self, mappings_dict):
        operations = self.parameters['operations']
        new_ops = []
        for op in operations:
            for key, value in mappings_dict.items():
                op = op.replace(key, value)
            new_ops.append(op)
        return new_ops

    def _get_associated_datasets(self):
        operations = self.parameters['operations']
        in_datasets = self.get_in_datasets()
        data_names = [d.get_name() for d in in_datasets]
        index = []
        for op in operations:
            names = [d for d in data_names if op.find(d) > -1]
            index.append(data_names.index(names[0]))
        return index



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: Remove zingers (caused by scattered X-rays hitting the CCD chip \
                            directly)
   :platform: Unix
   :synopsis: A plugin working in sinogram space to removes zingers
.. moduleauthor:: Nghia Vo <scientificsoftware@diamond.ac.uk>

"""
from savu.plugins.plugin import Plugin
from savu.plugins.driver.cpu_plugin import CpuPlugin
from savu.plugins.utils import register_plugin
import numpy as np


@register_plugin
class DezingerSinogramDeprecated(Plugin, CpuPlugin):
    """
    Method to remove scratches in the reconstructed image caused by zingers
    :u*param tolerance: Threshold for detecting zingers, greater is less \
    sensitive. Default: 0.08.

    """

    def __init__(self):
        super(DezingerSinogramDeprecated, self).__init__("DezingerSinogramDeprecated")

    def setup(self):
        in_dataset, out_dataset = self.get_datasets()
        out_dataset[0].create_dataset(in_dataset[0])
        in_pData, out_pData = self.get_plugin_datasets()
        in_pData[0].plugin_data_setup('SINOGRAM', 'single')
        out_pData[0].plugin_data_setup('SINOGRAM', 'single')

    def pre_process(self):
        in_pData = self.get_plugin_in_datasets()
        width_dim = in_pData[0].get_data_dimension_by_axis_label('detector_x')
        height_dim = \
            in_pData[0].get_data_dimension_by_axis_label('rotation_angle')
        sino_shape = list(in_pData[0].get_shape())
        self.width1 = sino_shape[width_dim]
        self.height1 = sino_shape[height_dim]

    def process_frames(self, data):
        tolerance = self.parameters['tolerance']
        nhigh = 1.0 + tolerance
        sinogram = np.copy(data[0])
        sinogram[sinogram == 0.0] = np.mean(sinogram)
        sinogram_d = np.roll(sinogram, 1, axis=0)
        sinogram_u = np.roll(sinogram, -1, axis=0)

        list_top = sinogram[0]
        list_top1 = sinogram[1]
        list_top2 = list_top / list_top1
        list_top[list_top2 > nhigh] = list_top1[list_top2 > nhigh]

        list_bottom = sinogram[-1]
        list_bottom1 = sinogram[-2]
        list_bottom2 = list_bottom / list_bottom1
        list_bottom[list_bottom2 > nhigh] = list_bottom1[list_bottom2 > nhigh]

        mat_ratio_d = sinogram / sinogram_d
        sinogram[mat_ratio_d > nhigh] = sinogram_d[mat_ratio_d > nhigh]
        mat_ratio_u = sinogram / sinogram_u
        sinogram[mat_ratio_u > nhigh] = sinogram_u[mat_ratio_u > nhigh]
        sinogram[0] = list_top
        sinogram[-1] = list_bottom
        return sinogram 



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# $Id: dezing_filter.py 467 2016-02-16 11:40:42Z kny48981 $


"""
.. module:: dezinger
   :platform: Unix
   :synopsis: A plugin to remove zingers

.. moduleauthor:: Mark Basham <scientificsoftware@diamond.ac.uk>

"""

import numpy as np
import dezing

from savu.plugins.filters.base_filter import BaseFilter
from savu.plugins.driver.cpu_plugin import CpuPlugin
from savu.plugins.utils import register_plugin


@register_plugin
class Dezinger(BaseFilter, CpuPlugin):
    """
    A plugin for cleaning x-ray strikes based on statistical evaluation of \
    the near neighbourhood
    
    :param outlier_mu: Threshold for defecting outliers, greater is less \
    sensitive. Default: 10.0.
    :param kernel_size: Number of frames included in average. Default: 5.
    :param mode: output mode, 0=normal 5=zinger strength 6=zinger \
        yes/no. Default: 0.
    """

    def __init__(self):
        super(Dezinger, self).__init__("Dezinger")
        self.warnflag = 0
        self.errflag = 0

    def pre_process(self):
        # Apply dezing to dark and flat images
        inData = self.get_in_datasets()[0]
        dark = inData.data.dark()
        flat = inData.data.flat()
        self.data_size = inData.get_shape()

        pad_list = ((self.pad, self.pad), (0, 0), (0, 0))

        # dezing the dark field        print "*****in data shape in base filter", in_dataset[0].get_shape()

        if dark.size:
            (retval, self.warnflag, self.errflag) = dezing.setup_size(
                dark.shape, self.parameters['outlier_mu'], self.pad,
                mode=self.parameters['mode'])
            dark = self._dezing(np.pad(dark, pad_list, mode='edge'))
            inData.data.update_dark(dark[self.pad:-self.pad])
            (retval, self.warnflag, self.errflag) = dezing.cleanup()

        # dezing the flat field
        if flat.size:
            (retval, self.warnflag, self.errflag) = dezing.setup_size(
                flat.shape, self.parameters['outlier_mu'],
                self.pad, mode=self.parameters['mode'])
            flat = self._dezing(np.pad(flat, pad_list, mode='edge'))
            inData.data.update_flat(flat[self.pad:-self.pad])
            (retval, self.warnflag, self.errflag) = dezing.cleanup()

        # setup dezing for data
        self._dezing_setup(self.data_size)

    def _dezing_setup(self, shape):
        (retval, self.warnflag, self.errflag) = \
            dezing.setup_size(shape, self.parameters['outlier_mu'],
                              self.pad, mode=self.parameters['mode'])

    def _process_calibration_frames(self, data):
        nSlices = data.shape[self.proj_dim] - 2*self.pad
        nSublists = int(np.ceil(nSlices/float(self.frame_limit)))
        idx = np.array_split(np.arange(self.pad, nSlices+self.pad), nSublists)
        idx = [np.arange(a[0]-self.pad, a[-1]+self.pad+1) for a in idx]
        out_sl = np.tile([slice(None)]*3, [len(idx), 1])
        out_sl[:, self.proj_dim] = idx
        result = np.empty_like(data)
        for sl in out_sl:
            result[tuple(sl)] = self._dezing(data[tuple(sl)])
        return result

    def _dezing(self, data):
        result = np.empty_like(data)
        (retval, self.warnflag, self.errflag) = dezing.run(data, result)
        return result

    def process_frames(self, data):
        return self._dezing(data[0])

    def post_process(self):
        (retval, self.warnflag, self.errflag) = dezing.cleanup()

    def get_max_frames(self):
        return 'multiple'

    def raw_data(self):
        return True

    def set_filter_padding(self, in_data, out_data):
        in_data = in_data[0]
        self.pad = (self.parameters['kernel_size'] - 1) // 2
        in_data.padding = {'pad_multi_frames': self.pad}
        out_data[0].padding = {'pad_multi_frames': self.pad}

    def executive_summary(self):
        if self.errflag != 0:
            return(["ERRORS detected in dezing plugin, Check the detailed \
log messages."])
        if self.warnflag != 0:
            return(["WARNINGS detected in dezing plugin, Check the detailed \
log messages."])
        return ["Nothing to Report"]



# Copyright 2019 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: Wrapper for CCPi-Regularisation Toolkit (GPU) for \
    efficient 2D/3D denoising
   :platform: Unix
   :synopsis: GPU modules of CCPi-Regularisation Toolkit (CcpiRegulToolkitCpu)

.. moduleauthor:: Daniil Kazantsev <scientificsoftware@diamond.ac.uk>
"""

from savu.plugins.plugin import Plugin
from savu.plugins.driver.multi_threaded_plugin import MultiThreadedPlugin
from savu.plugins.utils import register_plugin

from ccpi.filters.regularisers import ROF_TV, FGP_TV, SB_TV, PD_TV, LLT_ROF, TGV, NDF, Diff4th
from ccpi.filters.regularisers import PatchSelect, NLTV
from savu.data.plugin_list import CitationInformation

@register_plugin
class CcpiDenoisingCpu3d(Plugin, MultiThreadedPlugin):
    """
    'ROF_TV': Rudin-Osher-Fatemi Total Variation model;
    'FGP_TV': Fast Gradient Projection Total Variation model;
    'SB_TV': Split Bregman Total Variation model;
    'PD_TV': Primal-Dual Total variation model;
    'NLTV': Nonlocal Total Variation model;
    'TGV': Total Generalised Variation model;
    'LLT_ROF': Lysaker, Lundervold and Tai model combined with Rudin-Osher-Fatemi;
    'NDF': Nonlinear/Linear Diffusion model (Perona-Malik, Huber or Tukey);
    'DIFF4th': Fourth-order nonlinear diffusion model

    :param method: Choose methods |ROF_TV|FGP_TV|SB_TV|NLTV|TGV|LLT_ROF|NDF|Diff4th. Default: 'FGP_TV'.
    :param reg_par: Regularisation (smoothing) parameter. Default: 0.01.
    :param max_iterations: Total number of iterations. Default: 300.
    :param time_step: Time marching step, relevant for ROF_TV, LLT_ROF,\
     NDF, Diff4th methods. Default: 0.001.
    :param lipshitz_constant: TGV method, Lipshitz constant. Default: 12.
    :param alpha1: TGV method, parameter to control the 1st-order term. Default: 1.0.
    :param alpha0: TGV method, parameter to control the 2nd-order term. Default: 2.0.
    :param reg_parLLT: LLT-ROF method, parameter to control the 2nd-order term. Default: 0.05.
    :param penalty_type: NDF method, Penalty type for the duffison, choose from\
    huber, perona, tukey, constr, constrhuber. Default: 'huber'.
    :param edge_par: NDF and Diff4th methods, noise magnitude parameter. Default: 0.01.
    :param tolerance_constant: tolerance constant to stop iterations earlier. Default: 0.0.
    """

    def __init__(self):
        super(CcpiDenoisingCpu3d, self).__init__("CcpiDenoisingCpu3d")

    def setup(self):
        in_dataset, out_dataset = self.get_datasets()
        in_pData, out_pData = self.get_plugin_datasets()
        in_pData[0].plugin_data_setup('VOLUME_3D', 'single')
        out_dataset[0].create_dataset(in_dataset[0])
        out_pData[0].plugin_data_setup('VOLUME_3D', 'single')

    def pre_process(self):
        self.device = 'cpu'
        if (self.parameters['method'] == 'ROF_TV'):
            # set parameters for the ROF-TV method
            self.pars = {'algorithm': self.parameters['method'], \
                'regularisation_parameter':self.parameters['reg_par'],\
                'number_of_iterations': self.parameters['max_iterations'],\
                'time_marching_parameter': self.parameters['time_step'],\
                'tolerance_constant': self.parameters['tolerance_constant'] }
        if (self.parameters['method'] == 'FGP_TV'):
            # set parameters for the FGP-TV method
            self.pars = {'algorithm': self.parameters['method'], \
                'regularisation_parameter':self.parameters['reg_par'],\
                'number_of_iterations': self.parameters['max_iterations'],\
                'tolerance_constant': self.parameters['tolerance_constant'],\
                'methodTV': 0 ,\
                'nonneg': 0 }
        if (self.parameters['method'] == 'SB_TV'):
            # set parameters for the SB-TV method
            self.pars = {'algorithm': self.parameters['method'], \
                'regularisation_parameter':self.parameters['reg_par'],\
                'number_of_iterations': self.parameters['max_iterations'],\
                'tolerance_constant': self.parameters['tolerance_constant'],\
                'methodTV': 0 }
        if (self.parameters['method'] == 'TGV'):
            # set parameters for the TGV method
            self.pars = {'algorithm': self.parameters['method'],\
                'regularisation_parameter' : self.parameters['reg_par'],\
                'alpha1' : self.parameters['alpha1'],\
                'alpha0': self.parameters['alpha0'],\
                'number_of_iterations': self.parameters['max_iterations'],\
                'LipshitzConstant' :self.parameters['lipshitz_constant'],\
                'tolerance_constant': self.parameters['tolerance_constant']}
        if (self.parameters['method'] == 'LLT_ROF'):
            # set parameters for the LLT-ROF method
            self.pars = {'algorithm': self.parameters['method'], \
                'regularisation_parameter':self.parameters['reg_par'],\
                'regularisation_parameterLLT':self.parameters['reg_parLLT'], \
                'number_of_iterations': self.parameters['max_iterations'],\
                'time_marching_parameter': self.parameters['time_step'],\
                'tolerance_constant': self.parameters['tolerance_constant']}
        if (self.parameters['method'] == 'NDF'):
            # set parameters for the NDF method
            if (self.parameters['penalty_type'] == 'huber'):
                # Huber function for the diffusivity
                penaltyNDF = 1
            if (self.parameters['penalty_type'] == 'perona'):
                # Perona-Malik function for the diffusivity
                penaltyNDF = 2
            if (self.parameters['penalty_type'] == 'tukey'):
                # Tukey Biweight function for the diffusivity
                penaltyNDF = 3
            if (self.parameters['penalty_type'] == 'constr'):
                #  Threshold-constrained linear diffusion
                penaltyNDF = 4
            if (self.parameters['penalty_type'] == 'constrhuber'):
                #  Threshold-constrained huber diffusion
                penaltyNDF = 5                
            self.pars = {'algorithm': self.parameters['method'], \
                'regularisation_parameter':self.parameters['reg_par'],\
                'edge_parameter':self.parameters['edge_par'],\
                'number_of_iterations': self.parameters['max_iterations'],\
                'time_marching_parameter': self.parameters['time_step'],\
                'penalty_type': penaltyNDF,\
                'tolerance_constant': self.parameters['tolerance_constant']}
        if (self.parameters['method'] == 'Diff4th'):
            # set parameters for the Diff4th method
            self.pars = {'algorithm': self.parameters['method'], \
                'regularisation_parameter':self.parameters['reg_par'],\
                'edge_parameter':self.parameters['edge_par'],\
                'number_of_iterations': self.parameters['max_iterations'],\
                'time_marching_parameter': self.parameters['time_step'],\
                'tolerance_constant': self.parameters['tolerance_constant']}
        if (self.parameters['method'] == 'NLTV'):
            # set parameters for the NLTV method
            self.pars = {'algorithm': self.parameters['method'], \
                'regularisation_parameter':self.parameters['reg_par'],\
                'edge_parameter':self.parameters['edge_par'],\
                'number_of_iterations': self.parameters['max_iterations']}
        return self.pars

    def process_frames(self, data):
        import numpy as np
        input_temp = np.nan_to_num(data[0])
        input_temp[input_temp > 10**15] = 0.0
        self.pars['input'] = input_temp
        # Running Ccpi-RGLTK modules on GPU
        if (self.parameters['method'] == 'ROF_TV'):
            (im_res,infogpu) = ROF_TV(self.pars['input'],
                            self.pars['regularisation_parameter'],
                            self.pars['number_of_iterations'],
                            self.pars['time_marching_parameter'],
                            self.pars['tolerance_constant'],
                            self.device)
        if (self.parameters['method'] == 'FGP_TV'):
            (im_res,infogpu) = FGP_TV(self.pars['input'],
                            self.pars['regularisation_parameter'],
                            self.pars['number_of_iterations'],
                            self.pars['tolerance_constant'],
                            self.pars['methodTV'],
                            self.pars['nonneg'], self.device)
        if (self.parameters['method'] == 'SB_TV'):
            (im_res,infogpu) = SB_TV(self.pars['input'],
                            self.pars['regularisation_parameter'],
                            self.pars['number_of_iterations'],
                            self.pars['tolerance_constant'],
                            self.pars['methodTV'], self.device)
        if (self.parameters['method'] == 'TGV'):
            (im_res,infogpu) = TGV(self.pars['input'],
                            self.pars['regularisation_parameter'],
                            self.pars['alpha1'],
                            self.pars['alpha0'],
                            self.pars['number_of_iterations'],
                            self.pars['LipshitzConstant'],
                            self.pars['tolerance_constant'], self.device)
        if (self.parameters['method'] == 'LLT_ROF'):
            (im_res,infogpu) = LLT_ROF(self.pars['input'],
                           self.pars['regularisation_parameter'],
                           self.pars['regularisation_parameterLLT'],
                           self.pars['number_of_iterations'],
                           self.pars['time_marching_parameter'],
                           self.pars['tolerance_constant'], self.device)
        if (self.parameters['method'] == 'NDF'):
            (im_res,infogpu) = NDF(self.pars['input'],
         	           self.pars['regularisation_parameter'],
                           self.pars['edge_parameter'],
                           self.pars['number_of_iterations'],
                           self.pars['time_marching_parameter'],
                           self.pars['penalty_type'],
                           self.pars['tolerance_constant'], self.device)
        if (self.parameters['method'] == 'DIFF4th'):
            (im_res,infogpu) = Diff4th(self.pars['input'],
                           self.pars['regularisation_parameter'],
                           self.pars['edge_parameter'],
                           self.pars['number_of_iterations'],
                          self.pars['time_marching_parameter'],
                          self.pars['tolerance_constant'], self.device)
        if (self.parameters['method'] == 'NLTV'):
            pars_NLTV = {'algorithm' : PatchSelect, \
                         'input' : self.pars['input'],\
                         'searchwindow': 9, \
                         'patchwindow': 2,\
                         'neighbours' : 17 ,\
                         'edge_parameter':self.pars['edge_parameter']}
            H_i, H_j, Weights = PatchSelect(pars_NLTV['input'],
                                            pars_NLTV['searchwindow'],
                                            pars_NLTV['patchwindow'],
                                            pars_NLTV['neighbours'],
                                            pars_NLTV['edge_parameter'], self.device)
            parsNLTV_init = {'algorithm' : NLTV, \
                             'input' : pars_NLTV['input'],\
                             'H_i': H_i, \
                             'H_j': H_j,\
                             'H_k' : 0,\
                             'Weights' : Weights,\
                             'regularisation_parameter': self.pars['regularisation_parameter'],\
                             'iterations': self.pars['number_of_iterations']}
            im_res = NLTV(parsNLTV_init['input'],
                              parsNLTV_init['H_i'],
                              parsNLTV_init['H_j'],
                              parsNLTV_init['H_k'],
                              parsNLTV_init['Weights'],
                              parsNLTV_init['regularisation_parameter'],
                              parsNLTV_init['iterations'])
            del H_i,H_j,Weights
        #print "calling process frames", data[0].shape
        return im_res

    def nInput_datasets(self):
        return 1
    def nOutput_datasets(self):
        return 1

    def get_citation_information(self):
        cite_info1 = CitationInformation()
        cite_info1.name = 'citation1'
        cite_info1.description = \
            ("The CCPi-Regularisation toolkit provides a set of variational regularisers (denoisers) which can be embedded in a plug-and-play fashion into proximal splitting methods for image reconstruction. CCPi-RGL comes with algorithms that can satisfy various prior expectations of the reconstructed object, for example being piecewise-constant or piecewise-smooth nature.")
        cite_info1.bibtex = \
            ("@article{kazantsev2019,\n" +
             "title={CCPi-Regularisation Toolkit for computed tomographic image reconstruction with proximal splitting algorithms},\n" +
             "author={Daniil and Kazantsev, Edoardo and Pasca, Martin and Turner, Philip and Withers},\n" +
             "journal={Software X},\n" +
             "volume={9},\n" +
             "number={},\n" +
             "pages={317-323},\n" +
             "year={2019},\n" +
             "publisher={Elsevier}\n" +
             "}")
        cite_info1.endnote = \
            ("%0 Journal Article\n" +
             "%T CCPi-Regularisation Toolkit for computed tomographic image reconstruction with proximal splitting algorithms\n" +
             "%A Kazantsev, Daniil\n" +
             "%A Pasca, Edoardo\n" +
             "%A Turner, Martin\n" +
             "%A Withers, Philip\n" +
             "%J Software X\n" +
             "%V 9\n" +
             "%N \n" +
             "%P 317--323\n" +
             "%@ --\n" +
             "%D 2019\n" +
             "%I Elsevier\n")
        cite_info1.doi = "doi: 10.1016/j.softx.2019.04.003"

        if (self.parameters['method'] == 'ROF_TV'):
            cite_info2 = CitationInformation()
            cite_info2.name = 'citation2'
            cite_info2.description = ("Rudin-Osher-Fatemi explicit PDE minimisation method for smoothed Total Variation regulariser")
            cite_info2.bibtex = \
            ("@article{ROF1992,\n" +
             "title={Nonlinear total variation based noise removal algorithms},\n" +
             "author={L. Rudin, S. Osher, E. Fatemi},\n" +
             "journal={Physica D.},\n" +
             "volume={60},\n" +
             "number={},\n" +
             "pages={259--268},\n" +
             "year={1992},\n" +
             "publisher={Elsevier}\n" +
             "}")
            cite_info2.endnote = \
            ("%0 Journal Article\n" +
             "%T Nonlinear total variation based noise removal algorithms\n" +
             "%A Rudin, L.\n" +
             "%A Osher, S.\n" +
             "%A Fatemi, Fatemi\n" +
             "%J Physica D.\n" +
             "%V 60\n" +
             "%N \n" +
             "%P 259--268\n" +
             "%@ \n" +
             "%D 1992\n" +
             "%I Elsevier\n")
            cite_info2.doi = "doi: 10.1016/0167-2789(92)90242-F"
        if (self.parameters['method'] == 'FGP_TV'):
            cite_info2 = CitationInformation()
            cite_info2.name = 'citation2'
            cite_info2.description = ("Fast-Gradient-Projection algorithm for Total Variation regulariser")
            cite_info2.bibtex = \
            ("@article{FGP2009,\n" +
             "title={Fast gradient-based algorithms for constrained total variation image denoising and deblurring problems},\n" +
             "author={Amir and Beck, Mark and Teboulle},\n" +
             "journal={IEEE Transactions on Image Processing},\n" +
             "volume={18},\n" +
             "number={11},\n" +
             "pages={2419--2434},\n" +
             "year={2009},\n" +
             "publisher={IEEE}\n" +
             "}")
            cite_info2.endnote = \
            ("%0 Journal Article\n" +
             "%T Fast gradient-based algorithms for constrained total variation image denoising and deblurring problems\n" +
             "%A Beck, Amir\n" +
             "%A Teboulle, Mark.\n" +
             "%J IEEE Transactions on Image Processing\n" +
             "%V 18\n" +
             "%N 11\n" +
             "%P 2419--2434\n" +
             "%@ \n" +
             "%D 2009\n" +
             "%I IEEE\n")
            cite_info2.doi = "doi: 10.1109/TIP.2009.2028250"
        if (self.parameters['method'] == 'SB_TV'):
            cite_info2 = CitationInformation()
            cite_info2.name = 'citation2'
            cite_info2.description = ("The Split Bregman approach for Total Variation regulariser")
            cite_info2.bibtex = \
            ("@article{SBTV2009,\n" +
             "title={The split Bregman method for L1-regularized problems},\n" +
             "author={Tom and Goldstein, Stanley and Osher},\n" +
             "journal={SIAM journal on imaging sciences},\n" +
             "volume={2},\n" +
             "number={2},\n" +
             "pages={323--343},\n" +
             "year={2009},\n" +
             "publisher={SIAM}\n" +
             "}")
            cite_info2.endnote = \
            ("%0 Journal Article\n" +
             "%T The split Bregman method for L1-regularized problems\n" +
             "%A Goldstein, Tom\n" +
             "%A Osher, Stanley.\n" +
             "%J SIAM journal on imaging sciences\n" +
             "%V 2\n" +
             "%N 2\n" +
             "%P 323--343\n" +
             "%@ \n" +
             "%D 2009\n" +
             "%I SIAM\n")
            cite_info2.doi = "doi: 10.1137/080725891"
        if (self.parameters['method'] == 'TGV'):
            cite_info2 = CitationInformation()
            cite_info2.name = 'citation2'
            cite_info2.description = ("Total generalized variation regulariser for piecewise-smooth recovery")
            cite_info2.bibtex = \
            ("@article{TGV2010,\n" +
             "title={Total generalized variation},\n" +
             "author={Kristian and Bredies, Karl and Kunisch, Thomas and Pock},\n" +
             "journal={SIAM journal on imaging sciences},\n" +
             "volume={3},\n" +
             "number={3},\n" +
             "pages={492--526},\n" +
             "year={2010},\n" +
             "publisher={SIAM}\n" +
             "}")
            cite_info2.endnote = \
            ("%0 Journal Article\n" +
             "%T Total generalized variation\n" +
             "%A Bredies, Kristian\n" +
             "%A Kunisch, Kurl.\n" +
             "%A Pock, Thomas.\n" +
             "%J SIAM journal on imaging sciences\n" +
             "%V 3\n" +
             "%N 3\n" +
             "%P 492--526\n" +
             "%@ \n" +
             "%D 2010\n" +
             "%I SIAM\n")
            cite_info2.doi = "doi: 10.1137/080725891"
        if (self.parameters['method'] == 'LLT_ROF'):
            cite_info2 = CitationInformation()
            cite_info2.name = 'citation2'
            cite_info2.description = ("Combination for ROF model and LLT for piecewise-smooth recovery")
            cite_info2.bibtex = \
            ("@article{ROFLLT2017,\n" +
             "title={Model-based iterative reconstruction using higher-order regularization of dynamic synchrotron data},\n" +
             "author={Daniil and Kazantsev, Enyu and Guo, Andre and Phillion, Philip and Withers, Peter and Lee},\n" +
             "journal={Measurement Science and Technology},\n" +
             "volume={28},\n" +
             "number={9},\n" +
             "pages={094004},\n" +
             "year={2017},\n" +
             "publisher={IoP}\n" +
             "}")
            cite_info2.endnote = \
            ("%0 Journal Article\n" +
             "%T Model-based iterative reconstruction using higher-order regularization of dynamic synchrotron data\n" +
             "%A Kazantsev, Daniil\n" +
             "%A Enyu, Guo.\n" +
             "%A Phillion, Andre.\n" +
             "%A Withers, Philip.\n" +
             "%A Lee, Peter.\n" +
             "%J Measurement Science and Technology\n" +
             "%V 28\n" +
             "%N 9\n" +
             "%P 094004\n" +
             "%@ \n" +
             "%D 2017\n" +
             "%I IoP\n")
            cite_info2.doi = "doi: 10.1088/1361-6501"
        if (self.parameters['method'] == 'NDF'):
            cite_info2 = CitationInformation()
            cite_info2.name = 'citation2'
            cite_info2.description = ("Nonlinear or linear duffison as a regulariser ")
            cite_info2.bibtex = \
            ("@article{NDF1990,\n" +
             "title={Scale-space and edge detection using anisotropic diffusion},\n" +
             "author={Pietro and Perona, Jitendra and Malik},\n" +
             "journal={IEEE Transactions on pattern analysis and machine intelligence},\n" +
             "volume={12},\n" +
             "number={7},\n" +
             "pages={629--639},\n" +
             "year={1990},\n" +
             "publisher={IEEE}\n" +
             "}")
            cite_info2.endnote = \
            ("%0 Journal Article\n" +
             "%T Scale-space and edge detection using anisotropic diffusion\n" +
             "%A Pietro, Perona\n" +
             "%A Jitendra, Malik.\n" +
             "%J IEEE Transactions on pattern analysis and machine intelligence\n" +
             "%V 12\n" +
             "%N 7\n" +
             "%P 629--639\n" +
             "%@ \n" +
             "%D 1990\n" +
             "%I IEEE\n")
            cite_info2.doi = "doi: 10.1109/34.56205"
        if (self.parameters['method'] == 'DIFF4th'):
            cite_info2 = CitationInformation()
            cite_info2.name = 'citation2'
            cite_info2.description = ("Anisotropic diffusion of higher order for piecewise-smooth recovery")
            cite_info2.bibtex = \
            ("@article{ADF2011,\n" +
             "title={An anisotropic fourth-order diffusion filter for image noise removal},\n" +
             "author={Mohammad Reza and Hajiaboli},\n" +
             "journal={International Journal of Computer Vision},\n" +
             "volume={92},\n" +
             "number={2},\n" +
             "pages={177--191},\n" +
             "year={2011},\n" +
             "publisher={Springer}\n" +
             "}")
            cite_info2.endnote = \
            ("%0 Journal Article\n" +
             "%T Anisotropic diffusion of higher order for piecewise-smooth recovery\n" +
             "%A Mohammad Reza, Hajiaboli\n" +
             "%J International Journal of Computer Vision\n" +
             "%V 92\n" +
             "%N 2\n" +
             "%P 177--191\n" +
             "%@ \n" +
             "%D 2011\n" +
             "%I Springer\n")
            cite_info2.doi = "doi: 10.1007/s11263-010-0330-1"
        if (self.parameters['method'] == 'NLTV'):
            cite_info2 = CitationInformation()
            cite_info2.name = 'citation2'
            cite_info2.description = ("Nonlocal discrete regularization on weighted graphs: a framework for image and manifold processing")
            cite_info2.bibtex = \
            ("@article{abd2008,\n" +
             "title={Nonlocal discrete regularization on weighted graphs: a framework for image and manifold processing},\n" +
             "author={Abderrahim and Lezoray and Bougleux},\n" +
             "journal={IEEE Trans. Image Processing},\n" +
             "volume={17},\n" +
             "number={7},\n" +
             "pages={1047--1060},\n" +
             "year={2008},\n" +
             "publisher={IEEE}\n" +
             "}")
            cite_info2.endnote = \
            ("%0 Journal Article\n" +
             "%T Nonlocal discrete regularization on weighted graphs: a framework for image and manifold processing\n" +
             "%A Abderrahim, Lezoray, Bougleux\n" +
             "%J IEEE Trans. Image Processing\n" +
             "%V 17\n" +
             "%N 7\n" +
             "%P 1047--1060\n" +
             "%@ \n" +
             "%D 2008\n" +
             "%I IEEE\n")
            cite_info2.doi = "doi: 10.1109/TIP.2008.924284"
        return [cite_info1, cite_info2]



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: data_create
   :platform: Unix
   :synopsis: A class inherited by Data class that deals with data object \
   creation.

.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>

"""
import copy
import h5py
import numpy as np

from savu.core.utils import docstring_parameter
import savu.data.data_structures.data_notes as notes


class DataCreate(object):
    """ Class that deals with creating a data object.
    """

    def __init__(self, name='DataCreate'):
        self.dtype = None
        self.remove = False

    @docstring_parameter(notes._create.__doc__, notes._shape.__doc__,
                         notes.axis_labels.__doc__, notes.patterns.__doc__)
    def create_dataset(self, *args, **kwargs):
        """ Set up required information when an output dataset has been
        created by a plugin.

        :arg Data: A data object
        :keyword tuple shape: The shape of the dataset
        :keyword list axis_labels: The axis_labels associated with the datasets
        :keyword patterns: The patterns associated with the dataset (optional,\
            see note below)
        :keyword type dtype: Type of the data (optional: Defaults to \
            np.float32)
        :keyword bool remove: Remove from framework after completion \
        (no link in .nxs file) (optional: Defaults to False.)
        :keyword bool raw: Keep dark and flats (ImageKey or NoImageKey)

        {0} \n {1} \n {2} \n {3}

        """
        #self.dtype =
        self.set_dtype(kwargs.get('dtype', np.float32))
        self.remove = kwargs.get('remove', False)
        self.raw = kwargs.get('raw', False)
        self.transport = kwargs.get('transport', None)
        if self.transport:
            self.exp.meta_data.set('transport', self.transport)
            self._set_transport_data(self.transport)

        if len(args) is 1:
            self.__create_dataset_from_object(args[0])
        else:
            self.__create_dataset_from_kwargs(kwargs)
        self.get_preview().set_preview([])

    def set_dtype(self, dtype):
        if not dtype:
            if not self.data:
                plugin = self._get_plugin_data()._plugin.__class__.__name__
                raise Exception("Please create all output datasets before "
                                "setting plugin data in %s plugin.\n" % plugin)
            elif hasattr(self.data, 'dtype'):
                dtype = self.data.dtype
            else:
                h5 = h5py._hl.dataset.Dataset
                dtype = self.data.dtype if isinstance(self.data, h5) else \
                    self.data.data.dtype
        self.dtype = np.dtype(dtype)

    def get_dtype(self):
        return self.dtype

    def __create_dataset_from_object(self, data_obj):
        """ Create a dataset from an existing Data object.
        """
        patterns = copy.deepcopy(data_obj.get_data_patterns())
        self.__copy_labels(data_obj)
        self.__find_and_set_shape(data_obj)
        self._set_data_patterns(patterns)
        self.raw = data_obj.data if self.raw else None

    def __create_dataset_from_kwargs(self, kwargs):
        """ Create dataset from kwargs. """
        try:
            shape = kwargs['shape']
            self.__create_axis_labels(kwargs['axis_labels'])
        except KeyError:
            raise Exception("Please state axis_labels and shape when "
                            "creating a new dataset")
        if isinstance(shape, DataCreate):
            self.__find_and_set_shape(shape)
        else:
            pData = self._get_plugin_data()
            pData._set_shape_before_tuning(copy.copy(shape))
            self.set_shape(shape + tuple(pData.extra_dims))

        if 'patterns' in kwargs:
            patterns = self.__copy_patterns(kwargs['patterns'])
            self._set_data_patterns(patterns)

    def __copy_patterns(self, copy_data):
        """ Copy patterns """
        if isinstance(copy_data, DataCreate):
            patterns = copy.deepcopy(copy_data.get_data_patterns())
        else:
            data = list(copy_data.keys())[0]
            pattern_list = copy_data[data]

            all_patterns = copy.deepcopy(data.get_data_patterns())
            if len(pattern_list[0].split('.')) > 1:
                patterns = self.__copy_patterns_removing_dimensions(
                    pattern_list, all_patterns, len(data.get_shape()))
            else:
                patterns = {}
                for pattern in pattern_list:
                    patterns[pattern] = all_patterns[pattern]
        return patterns

    def __copy_patterns_removing_dimensions(self, pattern_list, all_patterns,
                                            nDims):
        """ Copy patterns but remove specified dimensions from them. """
        copy_patterns = {}
        for new_pattern in pattern_list:
            name, all_dims = new_pattern.split('.')
            if name is '*':
                copy_patterns = all_patterns
            else:
                copy_patterns[name] = all_patterns[name]
            dims = tuple(map(int, all_dims.split(',')))
            dims = self._non_negative_directions(dims, nDims=nDims)

        dim_map = [a for a in range(nDims) if a not in dims]
        patterns = {}
        for name, pattern_dict in copy_patterns.items():
            empty_flag = False
            for ddir in ['slice_dims', 'core_dims']:
                s_dims = self._non_negative_directions(
                    pattern_dict[ddir], nDims=nDims)
                new_dims = [sd for sd in s_dims if sd not in dims]
                new_dims = [dim_map.index(n) for n in new_dims]
                pattern_dict[ddir] = tuple(new_dims)
                if not new_dims:
                    empty_flag = True
            if empty_flag is False:
                patterns[name] = pattern_dict
        return patterns

    def __create_axis_labels(self, axis_labels):
        """ Create axis labels. """
        if isinstance(axis_labels, DataCreate):
            self.__copy_labels(axis_labels)
        elif isinstance(axis_labels, dict):
            data = list(axis_labels.keys())[0]
            self.__copy_labels(data)
            self.__amend_axis_labels(axis_labels[data])
        else:
            self.set_axis_labels(*axis_labels)
            # if parameter tuning
            if self._get_plugin_data().multi_params_dict:
                self.__add_extra_dims_labels()

    def __copy_labels(self, copy_data):
        """ Copy axis labels. """
        nDims = copy.copy(copy_data.data_info.get('nDims'))
        axis_labels = \
            copy.copy(copy_data.data_info.get('axis_labels'))
        self.data_info.set('nDims', nDims)
        self.data_info.set('axis_labels', axis_labels)
        # if parameter tuning
        if self._get_plugin_data().multi_params_dict:
            self.__add_extra_dims_labels()

    def __add_extra_dims_labels(self):
        """ Add axis labels to extra dimensions created by parameter tuning.
        """
        params_dict = self._get_plugin_data().multi_params_dict
        # add multi_params axis labels from dictionary in pData
        nDims = self.data_info.get('nDims')
        axis_labels = self.data_info.get('axis_labels')
        axis_labels.extend([0]*len(params_dict))
        for key, value in params_dict.items():
            title = value['label']
            name, unit = title.split('.')
            axis_labels[nDims + key] = {name: unit}
            # add parameter values to the meta_data
            self.meta_data.set(name, np.array(value['values']))
        self.data_info.set('nDims', nDims + len(self.extra_dims))
        self.data_info.set('axis_labels', axis_labels)

    def __amend_axis_labels(self, *args):
        """ Helper function to remove, replace/add or insert axis_labels into
        existing axis_labels
        """
        removed_dims = 0
        for arg in args[0]:
            if len(arg.split('.')) is 1:
                self.__remove_axis_labels(arg, removed_dims)
                removed_dims += 1
            else:
                if len(arg.split('~')) is 1:
                    self.__replace_axis_labels(arg)
                else:
                    self.__insert_axis_labels(arg)

    def __remove_axis_labels(self, label, removed_dims):
        """ Remove axis labels. """
        axis_labels = self.get_axis_labels()
        del axis_labels[int(label) - removed_dims]
        self.data_info.set(
            'nDims', self.data_info.get('nDims') - 1)

    def __replace_axis_labels(self, label):
        """ Replace or add axis labels. """
        axis_labels = self.get_axis_labels()
        label = label.split('.')
        axis_labels[int(label[0])] = {label[1]: label[2]}
        if int(label[0]) > self.data_info.get('nDims'):
            self.data_info.set(
                'nDims', self.data_info.get('nDims') + 1)

    def __insert_axis_labels(self, label):
        """ Insert axis labels. """
        axis_labels = self.get_axis_labels()
        label = label.split('~')[1].split('.')
        axis_labels.insert(int(label[0]), {label[1]: label[2]})
        self.data_info.set('nDims', self.data_info.get('nDims') + 1)

    def _set_data_patterns(self, patterns):
        """ Add missing dimensions to patterns and populate data info dict. """
        all_dims = list(range(len(self.get_shape())))
        for p in patterns:
            pDims = patterns[p]['core_dims'] + patterns[p]['slice_dims']
            for dim in all_dims:
                if dim not in pDims:
                    patterns[p]['slice_dims'] += (dim,)
        self.data_info.set('data_patterns', patterns)

    def __find_and_set_shape(self, data):
        """ Add any extra dimensions, from parameter tuning, to the shape and
        set the shape in the framework. """
        pData = self._get_plugin_data()
        pData._set_shape_before_tuning(copy.copy(data.get_shape()))
        new_shape = copy.copy(data.get_shape()) + tuple(pData.extra_dims)
        self.set_shape(new_shape)



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: multi_nxtomo_loader
   :platform: Unix
   :synopsis: A class for loading multiple standard tomography scans.

.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>

"""
import copy
import h5py
import tempfile
from os import path
import numpy as np

from savu.plugins.loaders.base_loader import BaseLoader
from savu.plugins.loaders.full_field_loaders.nxtomo_loader import NxtomoLoader
from savu.plugins.utils import register_plugin
from savu.data.data_structures.data_types.stitch_data import StitchData


@register_plugin
class MultiNxtomoLoader(BaseLoader):
    """
    A class to load multiple scans in Nexus format into one dataset.

    :param name: The name assigned to the dataset. Default: 'tomo'.
    :param file_name: The shared part of the name of each file\
        (not including .nxs). Default: None.
    :param data_path: Path to the data inside the \
        file. Default: 'entry1/tomo_entry/data/data'.
    :param dark: Optional path to the dark field data file, nxs path and \
        scale value. Default: [None, None, 1].
    :param flat: Optional Path to the flat field data file, nxs path and \
        scale value. Default: [None, None, 1]. 
    :param stack_or_cat: Stack or concatenate the data\
        (4D and 3D respectively). Default: 'stack'.
    :param stack_or_cat_dim: Dimension to stack or concatenate. Default: 3.
    :param axis_label: New axis label, if required, in the form\
        'name.units'. Default: 'scan.number'.
    :param range: The start and end of file numbers. Default: [0, 10].
    :param angles: A python statement to be evaluated or a file. Default: None.

    """

    def __init__(self, name='MultiNxtomoLoader'):
        super(MultiNxtomoLoader, self).__init__(name)

    def setup(self):
        nxtomo = self._get_nxtomo()
        preview = self.parameters['preview']
        stitch_dim = self.parameters['stack_or_cat_dim']
        nxtomo.parameters['preview'] = \
            [x for i, x in enumerate(preview) if i != stitch_dim]
        data_obj_list = self._get_data_objects(nxtomo)
        data_obj = \
            self.exp.create_data_object('in_data', self.parameters['name'])

        # dummy file
        filename = path.split('/')[-1] + '.h5'
        data_obj.backing_file = \
            h5py.File(tempfile.mkdtemp() + '/' + filename, 'a')

        stack_or_cat = self.parameters['stack_or_cat']
        data_obj.data = StitchData(data_obj_list, stack_or_cat, stitch_dim)

        if stack_or_cat == 'cat':
            nxtomo._setup_3d(data_obj)
            self._extend_axis_label_values(data_obj_list, data_obj)
        else:
            self._setup_4d(data_obj)

        data_obj.set_original_shape(data_obj.data.get_shape())
        self.set_data_reduction_params(data_obj)
        # Must do this here after preview has been applied
        if stack_or_cat == 'stack':
            self._set_nD_rotation_angle(data_obj_list, data_obj)
    
    def _get_nxtomo(self):
        nxtomo = NxtomoLoader()
        nxtomo.exp = self.exp
        nxtomo._populate_default_parameters()
        
        # update nxtomo parameters with any common keys
        shared_keys = set(nxtomo.parameters.keys()).intersection(
            set(self.parameters.keys()))
        for key in shared_keys:
            nxtomo.parameters[key] = self.parameters[key]
        return nxtomo    

    def _get_data_objects(self, nxtomo):
        rrange = self.parameters['range']
        file_list = list(range(rrange[0], rrange[1]+1))
        file_path = copy.copy(self.exp.meta_data.get('data_file'))
        file_name = '' if self.parameters['file_name'] is None else\
            self.parameters['file_name']

        data_obj_list = []
        for i in file_list:
            this_file = file_path + file_name + str(i) + '.nxs'
            self.exp.meta_data.set('data_file', this_file)
            nxtomo.setup()
            data_obj_list.append(self.exp.index['in_data']['tomo'])
            self.exp.index['in_data'] = {}

        self.exp.meta_data.set('data_file', file_path)
        return data_obj_list

    def _setup_4d(self, data_obj):
        axis_labels = \
            ['rotation_angle.degrees', 'detector_y.pixel', 'detector_x.pixel']

        extra_label = self.parameters['axis_label']
        axis_labels.append(extra_label)

        rot = axis_labels.index('rotation_angle.degrees')
        detY = axis_labels.index('detector_y.pixel')
        detX = axis_labels.index('detector_x.pixel')
        extra = axis_labels.index(extra_label)

        data_obj.set_axis_labels(*axis_labels)

        data_obj.add_pattern('PROJECTION', core_dims=(detX, detY),
                             slice_dims=(rot, extra))
        data_obj.add_pattern('SINOGRAM', core_dims=(detX, rot),
                             slice_dims=(detY, extra))

        data_obj.add_pattern('PROJECTION_STACK', core_dims=(detX, detY),
                             slice_dims=(extra, rot))
        data_obj.add_pattern('SINOGRAM_STACK', core_dims=(detX, rot),
                             slice_dims=(extra, detY))

    def _extend_axis_label_values(self, data_obj_list, data_obj):
        dim = self.parameters['stack_or_cat_dim']
        axis_name = list(data_obj.get_axis_labels()[dim].keys())[0].split('.')[0]

        new_values = np.zeros(data_obj.data.get_shape()[dim])
        inc = len(data_obj_list[0].meta_data.get(axis_name))

        for i in range(len(data_obj_list)):
            new_values[i*inc:i*inc+inc] = \
                data_obj_list[i].meta_data.get(axis_name)

        data_obj.meta_data.set(axis_name, new_values)

    def _set_nD_rotation_angle(self, data_obj_list, data_obj):
        shape = data_obj.get_shape()
        rot_dim_len = data_obj.get_shape()[
            data_obj.get_data_dimension_by_axis_label('rotation_angle')]
        new_values = np.zeros([rot_dim_len, len(data_obj_list)])
        for i in range(len(data_obj_list)):
            new_values[:, i] = \
                data_obj_list[i].meta_data.get('rotation_angle')
        data_obj.meta_data.set('rotation_angle', new_values)


    def get_dark_flat_slice_list(self, data_obj):
        slice_list = data_obj._preview._get_preview_slice_list()
        detX_dim = data_obj.get_data_dimension_by_axis_label('detector_x')
        detY_dim = data_obj.get_data_dimension_by_axis_label('detector_y')
        dims = list(set([detX_dim, detY_dim]))
        new_slice_list = []
        for d in dims:
            new_slice_list.append(slice_list[d])
        return new_slice_list



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: scikitimage_sart
   :platform: Unix
   :synopsis: Wrapper for scikitimage SART function

"""

import logging
from savu.plugins.reconstructions.base_recon import BaseRecon
from savu.data.plugin_list import CitationInformation
from savu.plugins.driver.cpu_plugin import CpuPlugin

import skimage.transform as transform
import numpy as np
from scipy import ndimage

from savu.plugins.utils import register_plugin


@register_plugin
class ScikitimageSart(BaseRecon, CpuPlugin):
    """
    A Plugin to reconstruct an image by filter back projection using the \
    inverse radon transform from scikit-image.

    :param iterations: Number of iterations in the reconstruction. Default: 1.
    :param output_size: Number of rows and columns in the \
        reconstruction. Default: None.
    :param filter: Filter used in frequency domain \
        filtering. Ramp filter used by default. Filters available: ramp, \
        shepp-logan, cosine, hamming, hann.  Assign None to use no \
        filter. Default: 'ramp'.
    :param interpolation: interpolation method used in reconstruction. \
        Methods available: 'linear', 'nearest', and 'cubic' \
        ('cubic' is slow). Default: 'linear'.
    :param circle: Assume the reconstructed image is zero outside the \
        inscribed circle. Also changes the default output_size to match the \
        behaviour of radon called with circle=True. Default: False.
    :param image: 2D array, dtype=float, optional.  Image containing an \
        initial reconstruction estimate. Shape of this array should be \
        (radon_image.shape[0], radon_image.shape[0]). The default is a filter \
        backprojection using scikit.image.iradon as "result". Default: None.

    :param projection_shifts : 1D array, dtype=float. Shift the projections \
        contained in radon_image (the sinogram) by this many pixels before \
        reconstructing the image. The i'th value defines the shift of the \
        i'th column of radon_image.  Default: None.
    :param clip : length-2 sequence of floats. Force all values in the \
        reconstructed tomogram to lie in the range \
        [clip[0], clip[1]]. Default: None.
    :param relaxation : float. Relaxation parameter for the update step. A \
        higher value can improve the convergence rate, but one runs the risk \
        of instabilities. Values close to or higher than 1 are not \
        recommended. Default: None.

    :~param outer_pad: Not required. Default: False.
    :~param centre_pad: Not required. Default: False.

    """

    def __init__(self):
        logging.debug("initialising Scikitimage SART")
        logging.debug("Calling super to make sure that all superclasses are " +
                      " initialised")
        super(ScikitimageSart, self).__init__("ScikitimageSart")

    def _shift(self, sinogram, centre_of_rotation):
        centre_of_rotation_shift = (sinogram.shape[0] // 2) - float(centre_of_rotation)
        return ndimage.interpolation.shift(sinogram, centre_of_rotation_shift)

    def process_frames(self, data):
        sino = data[0]
        centre_of_rotations, angles, vol_shape, init = self.get_frame_params()
        in_pData = self.get_plugin_in_datasets()[0]
        sinogram = np.swapaxes(sino, 0, 1)
        sinogram = self._shift(sinogram, centre_of_rotations)
        sino = sinogram.astype(np.float64)
        theta = np.linspace(0, 180, sinogram.shape[1])
        result = \
            transform.iradon(sino, theta=theta,
                             output_size=(in_pData.get_shape()[1]),
                             # self.parameters['output_size'],
                             filter='ramp',  # self.parameters['filter'],
                             interpolation='linear',
                             # self.parameters['linear'],
                             circle=False)  # self.parameters[False])

        for i in range(self.parameters["iterations"]):
            print("Iteration %i" % i)
            result = transform.iradon_sart(sino, theta=theta, image=result,
                                           # self.parameters['result'],
                                           projection_shifts=None,
                                           # self.parameters['None'],
                                           clip=None,
                                           # self.parameters[None],
                                           relaxation=0.15
                                           # self.parameters[0.15])
                                           )
        return result

    def get_max_frames(self):
        return 'single'

    def get_citation_information(self):
        cite_info = CitationInformation()
        cite_info.description = \
            ("The Tomographic reconstruction performed in this processing " +
             "chain is derived from this work.")
        cite_info.bibtex = \
            ("@book{avinash2001principles,\n" +
             " title={Principles of computerized tomographic imaging},\n" +
             " author={Kak, Avinash C. and Slaney, Malcolm},\n" +
             " year={2001},\n" +
             " publisher={Society for Industrial and Applied Mathematics}\n" +
             "}")
        cite_info.endnote = \
            ("%0 Book\n" +
             "%T Principles of computerized tomographic imaging\n" +
             "%A Kak, Avinash C.\n" +
             "%A Slaney, Malcolm\n" +
             "%@ 089871494X\n" +
             "%D 2001\n" +
             "%I Society for Industrial and Applied Mathematics")
        cite_info.doi = "https://doi.org/10.1137/1.9780898719277"
        return cite_info



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: nxstxm_loader
   :platform: Unix
   :synopsis: A class for loading nxstxm data

.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>

"""

from savu.plugins.loaders.mapping_loaders.base_multi_modal_loader \
    import BaseMultiModalLoader

from savu.plugins.utils import register_plugin


@register_plugin
class NxstxmLoader(BaseMultiModalLoader):
    """
    A class to load tomography data from an NXstxm file
    :param name: The name assigned to the dataset. Default: 'stxm'.
    """

    def __init__(self, name='NxstxmLoader'):
        super(NxstxmLoader, self).__init__(name)

    def setup(self):
        path = 'instrument/detector/data'
        data_obj, stxm_entry = \
            self.multi_modal_setup('NXstxm', path, self.parameters['name'])

        mono_energy = data_obj.backing_file[
            stxm_entry.name + '/instrument/monochromator/energy']
        self.exp.meta_data.set("mono_energy", mono_energy)

        self.set_data_reduction_params(data_obj)



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: tiff_saver
   :platform: Unix
   :synopsis: A class to save output in tiff format

.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>

"""

from mpi4py import MPI
import os
import numpy as np
from savu.plugins.savers.base_saver import BaseSaver
from savu.plugins.utils import register_plugin
from savu.plugins.driver.cpu_plugin import CpuPlugin
from fabio.edfimage import EdfImage
import collections


@register_plugin
class XrfSaver(BaseSaver, CpuPlugin):
    """
    A class to save xrf data to Diamond Mapping .nxs format. It can also save to edf and rgb for pymca visualisation.
    """

    def __init__(self, name='XrfSaver'):
        super(XrfSaver, self).__init__(name)
        self.folder = None
        self.data_name = None
        self.file_name = None
        self.group_name = None

    def pre_process(self):
        self.data_name = self.get_in_datasets()[0].get_name()
        self.group_name = self._get_group_name(self.data_name)
        self.folder = "%s/%s-%s" % (self.exp.meta_data.get("out_path"),
                                    self.name, self.data_name)
        if MPI.COMM_WORLD.rank == 0:
            if not os.path.exists(self.folder):
                os.makedirs(self.folder)
        in_pData, __out_pData = self.get_plugin_datasets()
        in_datasets, __out_datasets = self.get_datasets()
        self.axis_labels = in_datasets[0].get_axis_labels()
        sd = in_datasets[0].get_slice_dimensions()
        self.axes = [self.axis_labels[ix] for ix in sd]
        self.axis_info = []
        k = 0
        for ix in self.axes:
            axis_name = list(ix.keys())[0]
            try:
                foo = in_datasets[0].meta_data.get(axis_name)# take the first key
            except KeyError:
                foo = np.array(list(range(in_datasets[0].get_shape()[sd[k]]))) # if it doesn't exis then replace it with a range
            k+=1
            self.axis_info.append(foo)
        #self.axis_info = [in_datasets[0].meta_data.get(ix) for ix in self.axes]

    def process_frames(self, data):
        d=data[0] # first dataset in the list
        print("################################")
        print("number of datasets in list",len(data))
        print("shape of data[0]", data[0].shape)
        print("################################")
#         frame = self.get_current_slice_list()[0] # slice list for the first dataset
#         frame = [ix.start for ix in frame] #  convert to just numbers
#         channel = self.axis_info[0][frame[-2]]
#         elements = self.axis_info[1][frame[-1]]
#         foo = data[0]
#         header = collections.OrderedDict()     
#         header['HeaderID'] = 'EH:000001:000000:000000'
#         header['Image'] = '1'
#         header['ByteOrder'] = 'LowByteFirst'
#         header['DataType']=  'DoubleValue'
#         header['Dim_1'] = str(foo.shape[-1]) 
#         header['Dim 2'] = str(foo.shape[-2])
#         header['Size'] = str(8.0*np.prod(foo.shape))
#         out_title = "%s_channel_%s" % (elements.replace(" ","_"), str(channel))
#         header['Title'] = out_title
#         fout = EdfImage(foo, header)
#         filename = self.folder+os.sep + str(out_title) + '.edf'
#         fout.write(filename)

    def get_pattern(self):
        return "PROJECTION"


    def get_max_frames(self):
        in_datasets, __out_datasets = self.get_datasets()
        sh = in_datasets[0].get_shape()
        print("full dataset shape:",sh)
        nframes = np.prod(sh[-2:])# product of the channel and the number of elements in the fit
        print("max frames requested:",nframes, type(nframes))
        return nframes


# Copyright 2015 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
.. module:: hdf5_transport
   :platform: Unix
   :synopsis: Transport specific plugin list runner, passes the data to and \
       from the plugin.

.. moduleauthor:: Mark Basham <scientificsoftware@diamond.ac.uk>

"""

import os
import logging

from savu.core.transport_setup import MPI_setup
from savu.plugins.savers.utils.hdf5_utils import Hdf5Utils
from savu.core.transports.base_transport import BaseTransport


class Hdf5Transport(BaseTransport):

    def __init__(self):
        super(Hdf5Transport, self).__init__()
        os.environ['savu_mode'] = 'hdf5'
        self.count = 0

    def _transport_initialise(self, options):
        MPI_setup(options)
        self.exp_coll = None
        self.data_flow = []
        self.files = []

    def _transport_update_plugin_list(self):
        plugin_list = self.exp.meta_data.plugin_list
        saver_idx = plugin_list._get_savers_index()
        remove = []

        # check the saver plugin and turn off if it is hdf5
        for idx in saver_idx:
            if plugin_list.plugin_list[idx]['name'] == 'Hdf5Saver':
                remove.append(idx)
        for idx in sorted(remove, reverse=True):
            plugin_list._remove(idx)

    def _transport_pre_plugin_list_run(self):
        # run through the experiment (no processing) and create output files
        self.hdf5 = Hdf5Utils(self.exp)
        self.exp_coll = self.exp._get_collection()
        self.data_flow = self.exp.meta_data.plugin_list._get_dataset_flow()
        n_plugins = list(range(len(self.exp_coll['datasets'])))

        for i in n_plugins:
            self.exp._set_experiment_for_current_plugin(i)
            self.files.append(
                self._get_filenames(self.exp_coll['plugin_dict'][i]))
            self._set_file_details(self.files[i])
            self._setup_h5_files()  # creates the hdf5 files

    def _transport_pre_plugin(self):
        count = self.exp.meta_data.get('nPlugin')
        self._set_file_details(self.files[count])

    def _transport_post_plugin(self):
        for data in list(self.exp.index['out_data'].values()):
            if not data.remove:
                msg = self.__class__.__name__ + "_transport_post_plugin."
                self.exp._barrier(msg=msg)
                if self.exp.meta_data.get('process') == \
                        len(self.exp.meta_data.get('processes'))-1:
                    self._populate_nexus_file(data)
                    self.hdf5._link_datafile_to_nexus_file(data)
                self.exp._barrier(msg=msg)
                # reopen file as read-only
                self.hdf5._reopen_file(data, 'r')

    def _transport_terminate_dataset(self, data):
        self.hdf5._close_file(data)

    def _transport_checkpoint(self):
        """ The framework has determined it is time to checkpoint.  What
        should this transport mechanism do?"""
        cp = self.exp.checkpoint
        with self.hdf5._open_backing_h5(cp._file, 'a', mpi=False) as f:
            self._metadata_dump(f, 'in_data')
            self._metadata_dump(f, 'out_data')

    def _metadata_dump(self, f, gname):
        if gname in list(f.keys()):
            del f[gname]

        for data in list(self.exp.index[gname].values()):
            name = data.get_name()
            entry = f.require_group(gname + '/' + name)
            self._output_metadata(data, entry, name, dump=True)

    def _transport_kill_signal(self):
        """ An opportunity to send a kill signal to the framework.  Return
        True or False. """
        path = self.exp.meta_data.get('out_path')
        killsignal = os.path.join(path, 'killsignal')
        # jump to the end of the plugin run!

        if os.path.exists(killsignal):
            self.exp.meta_data.set('killsignal', True)
            logging.debug("***************** killsignal sent ****************")
            return True
        return False

    def _transport_cleanup(self, i):
        """ Any remaining cleanup after kill signal sent """
        n_plugins = len(self.exp_coll['datasets'])
        for i in range(i, n_plugins):
            self.exp._set_experiment_for_current_plugin(i)
            for data in list(self.exp.index['out_data'].values()):
                self.hdf5._close_file(data)




# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: data_test
   :platform: Unix
   :synopsis: unittest test class data structures

.. moduleauthor:: Mark Basham <scientificsoftware@diamond.ac.uk>

"""

import unittest

import savu.test.test_utils as tu
from savu.core.plugin_runner import PluginRunner


class Test(unittest.TestCase):

    def test_create_smaller_data_block(self):
        options = tu.set_experiment('tomo')
        plugin = 'savu.plugins.reshape.downsample_filter'
        tu.set_plugin_list(options, plugin)

        out_data_name = options['plugin_list'][1]['data']['out_datasets'][0]
        plugin_runner = PluginRunner(options)
        exp = plugin_runner._run_plugin_list()

        self.assertEqual(exp.index['in_data'][out_data_name].get_shape(),
                         (91, 68, 80))

if __name__ == "__main__":
    unittest.main()



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: savu_config.py
   :platform: Unix
   :synopsis: A command line tool for creating Savu plugin lists

.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>

"""



import re
import sys


import warnings
with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    from .content import Content
    from .completer import Completer
    from .display_formatter import ListDisplay, DispDisplay
    from . import arg_parsers as parsers
    from savu.plugins import utils as pu
    from . import config_utils as utils
    from .config_utils import parse_args
    from .config_utils import error_catcher


def _help(content, args):
    """ Display the help information"""
    print('%s Savu configurator commands %s\n' % tuple(['*'*21]*2))
    for key in list(commands.keys()):
        doc = commands[key].__doc__
        if doc:
            print("%8s : %s" % (key, commands[key].__doc__))
    print("\n%s\n* For more information about individual commands type "
          "'<command> -h' *\n%s\n" % tuple(['*'*70]*2))
    return content


@parse_args
@error_catcher
def _open(content, args):
    """ Open an existing process list."""
    content.fopen(args.file, update=True, skip=args.skip)
    _ref(content, '* -n')
    _disp(content, '-q')
    return content


@parse_args
@error_catcher
def _disp(content, args):
    """ Display the plugins in the current list."""
    range_dict = utils.__get_start_stop(content, args.start, args.stop)
    formatter = DispDisplay(content.plugin_list)
    verbosity = parsers._get_verbosity(args)
    level = 'all' if args.all else content.disp_level
    content.display(formatter, level=level, verbose=verbosity, **range_dict)
    return content


@parse_args
@error_catcher
def _list(content, args):
    """ List the available plugins. """
    list_content = Content()
    utils._populate_plugin_list(list_content, pfilter=args.string)
    if not len(list_content.plugin_list.plugin_list):
        print("No result found.")
        return content
    formatter = ListDisplay(list_content.plugin_list)
    verbosity = parsers._get_verbosity(args)
    list_content.display(formatter, verbose=verbosity)
    return content


@parse_args
@error_catcher
def _save(content, args):
    """ Save the current process list to file."""
    out_file = content.filename if args.input else args.filepath
    content.check_file(out_file)
    print()
    DispDisplay(content.plugin_list)._notices()
    content.save(out_file, check=input("Are you sure you want to save the "
                 "current data to %s' [y/N]" % (out_file)),
                 template=args.template)
    return content


@parse_args
@error_catcher
def _mod(content, args):
    """ Modify plugin parameters. """
    pos_str, subelem = args.param.split('.')
    content.modify(pos_str, subelem, ' '.join(args.value))
    _disp(content, pos_str)
    return content


@parse_args
@error_catcher
def _set(content, args):
    """ Set the status of the plugin to be on or off. """
    for pos in args.plugin_pos:
        content.on_and_off(pos, args.status.upper())
    _disp(content, '-q')
    return content


@parse_args
@error_catcher
def _add(content, args):
    """ Add a plugin to the list. """
    elems = content.get_positions()
    final = str(int(re.findall(r'\d+', elems[-1])[0])+1) if elems else 1
    content.add(args.name, args.pos if args.pos else str(final))
    _disp(content, '-q')
    return content


@parse_args
@error_catcher
def _ref(content, args):
    """ Refresh a plugin (update it). """
    positions = content.get_positions() if args.pos == ['*'] else args.pos
    for pos_str in positions:
        content.refresh(pos_str, defaults=args.defaults)
        if not args.nodisp:
            _disp(content, pos_str)
    return content


@parse_args
@error_catcher
def _rem(content, args):
    """ Remove plugin(s) from the list. """
    content.remove(content.find_position(args.pos))
    _disp(content, '-q')
    return content


@parse_args
@error_catcher
def _move(content, args):
    """ Move a plugin to a different position in the list."""
    content.move(args.orig_pos, args.new_pos)
    _disp(content, '-q')
    return content


@parse_args
@error_catcher
def _coll(content, arg):
    """ List all plugin collections. """
    colls = Completer([])._get_collections()
    print('\n')
    for c in colls:
        print('%s\n %s' % ('-'*40, c))
    print('-'*40, '\n')
    return content


def _clear(content, arg):
    """ Clear the current plugin list."""
    content.clear(check=input("Are you sure you want to clear the current "
                  "plugin list? [y/N]"))
    return content


def _exit(content, arg):
    """ Close the program."""
    content.set_finished(check=input("Are you sure? [y/N]"))
    return content


def _history(content, arg):
    hlen = utils.readline.get_current_history_length()
    for i in range(hlen):
        print("%5i : %s" % (i, utils.readline.get_history_item(i)))
    return content


commands = {'open': _open,
            'help': _help,
            'disp': _disp,
            'list': _list,
            'save': _save,
            'mod': _mod,
            'set': _set,
            'add': _add,
            'rem': _rem,
            'move': _move,
            'ref': _ref,
            'coll': _coll,
            'clear': _clear,
            'exit': _exit,
            'history': _history}


def main(test=False):
    """
    :param test: If test is True the last argument from sys.argv is removed,
                 as it contains the directory of the test scripts, which fails the
                 parsing of the arguments as it has an unexpected argument.

                 If test is False then nothing is touched.
    """

    print("Running the configurator")
    # required for running the tests locally or on travis
    # drops the last argument from pytest which is the test file/module
    if test:
        try:
            # find where the /scripts argument is
            index_of_scripts_argument = ["scripts" in arg for arg in sys.argv].index(True)
            # remove it, including every arguments after it (e.g --cov)
            sys.argv = sys.argv[:index_of_scripts_argument]
        except ValueError:
            # scripts was not part of the arguments passed in by the test
            pass

    args = parsers._config_arg_parser()
    if args.error:
        utils.error_level = 1

    print("Starting Savu Config tool (please wait for prompt)")

    _reduce_logging_level()

    content = Content(level="all" if args.disp_all else 'user')

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        # imports all the (working) plugin modules
        content.failed = utils.populate_plugins(error_mode=args.error,
                                                examples=args.examples)

    comp = Completer(commands=commands, plugin_list=pu.plugins)
    utils._set_readline(comp.complete)


    # if file flag is passed then open it here
    if args.file:
        commands['open'](content, args.file)

    print("\n*** Press Enter for a list of available commands. ***\n")

    utils.load_history_file(utils.histfile)

    while True:
        try:
            in_list = input(">>> ").strip().split(' ', 1)
        except KeyboardInterrupt:
            print()
            continue
        except EOFError:
            # makes possible exiting on CTRL + D (EOF, like Python interpreter)
            break

        command, arg = in_list if len(in_list) == 2 else in_list+['']
        command = command if command else 'help'
        if command not in commands:
            print("I'm sorry, that's not a command I recognise. Press Enter "
                  "for a list of available commands.")
        else:
            content = commands[command](content, arg)

        if content.is_finished():
            break

    print("Thanks for using the application")

def _reduce_logging_level():
    import logging
    logging.getLogger().setLevel(logging.CRITICAL)

if __name__ == '__main__':
    main()



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: mm_loader
   :platform: Unix
   :synopsis: A class for loading multi-modal data

.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>

"""

from savu.plugins.loaders.base_loader import BaseLoader
from savu.plugins.loaders.mapping_loaders.nxfluo_loader \
    import NxfluoLoader as fluo
from savu.plugins.loaders.mapping_loaders.nxxrd_loader \
    import NxxrdLoader as xrd
from savu.plugins.loaders.mapping_loaders.nxstxm_loader \
    import NxstxmLoader as stxm
from savu.plugins.loaders.mapping_loaders.nxmonitor_loader \
    import NxmonitorLoader as mon
import logging

from savu.plugins.utils import register_plugin
from savu.core.utils import docstring_parameter
from savu.plugins.loaders.mapping_loaders.base_multi_modal_loader \
    import BaseMultiModalLoader


@register_plugin
class MmLoader(BaseLoader):

    def __init__(self, name='MmLoader'):
        super(MmLoader, self).__init__(name)
        base = BaseMultiModalLoader()
        base._populate_default_parameters()
        self.doc_string = base.__doc__
        self.dict = base.parameters
        self.fluo_keys = self.set_params(fluo(), 'fluo')
        self.xrd_keys = self.set_params(xrd(), 'xrd')
        self.stxm_keys = self.set_params(stxm(), 'stxm')
        self.mon_keys = self.set_params(mon(), 'monitor')
        for key, value in self.dict.items():
            self.parameters[key] = value

    def set_params(self, inst, name):
        inst._populate_default_parameters()
        copy_keys = inst.parameters.keys() - self.dict.keys()
        for key in copy_keys:
            self.parameters[key] = inst.parameters[key]
        return list(copy_keys)

    def separate_params(self, name, keys):
        all_keys = list(self.dict.keys()) + keys
        new_dict = {}
        for key in [k for k in all_keys if k != 'name']:
            new_dict[key] = self.parameters[key]
        new_dict['name'] = self.name_dict[name]
        new_dict['preview'] = self.parameters['preview'][new_dict['name']]
        return new_dict

    @docstring_parameter(BaseMultiModalLoader.__doc__, xrd.__doc__,
                         stxm.__doc__, mon.__doc__, fluo.__doc__)
    def _override_class_docstring(self):
        """
        :u*param dataset_names: The names assigned to each dataset in the \
            order: fluorescence, diffraction, absorption, \
            monitor. Default: ['fluo', 'xrd', 'stxm', 'monitor'].
        :u*param preview: A slice list of required frames to apply to ALL \
        datasets, else a dictionary of slice lists where the key is the \
        dataset name. Default: [].

        {0} \n {1} \n {2} \n {3} \n {4}

        """
        pass

    def __set_names(self):
        fluo, xrd, stxm, monitor = self.parameters['dataset_names']
        self.name_dict = {'fluo': fluo, 'xrd': xrd, 'stxm': stxm,
                          'monitor': monitor}

    def __set_preview(self):
        preview_dict = {}
        preview = self.parameters['preview']
        for name in list(self.name_dict.values()):
            preview_dict[name] = preview[name] if isinstance(preview, dict) \
                and name in list(preview.keys()) else preview if \
                    isinstance(preview, list) else []
        self.parameters['preview'] = preview_dict

    def setup(self):
        self.__set_names()
        self.__set_preview()
        self._data_loader(fluo(), 'fluo', self.fluo_keys)
        self._data_loader(xrd(), 'xrd', self.xrd_keys)
        self._data_loader(stxm(), 'stxm', self.stxm_keys)
        self._data_loader(mon(), 'monitor', self.mon_keys)

    def _data_loader(self, inst, name, key):
        debug_str = 'This file contains an ' + name
        warn_str = 'This file does not contain a ' + name
        try:
            params = self.separate_params(name, key)
            inst.initialise(params, self.exp)
            logging.debug(debug_str)
        except IndexError:
            # Delete the data object if it has already been created.
            if name in self.exp.index['in_data']:
                del self.exp.index['in_data'][name]
            logging.warning(warn_str)
        except:
            raise

    def final_parameter_updates(self):
        # names of individual datasets are not required
        self.delete_parameter_entry('name')



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: camera_rot_correction
   :platform: Unix
   :synopsis: A plugin to apply a rotation to projection images
.. moduleauthor:: Malte Storm<malte.storm@diamond.ac.uk>
"""


from savu.plugins.filters.base_filter import BaseFilter
from savu.plugins.driver.cpu_plugin import CpuPlugin
from savu.plugins.utils import register_plugin

from skimage.transform import rotate
import numpy as np


@register_plugin
class CameraRotCorrection(BaseFilter, CpuPlugin):
    """
    A plugin to apply a rotation to projection images, for example to correct\
     for missing camera alignment.

    :u*param angle: The rotation angle for the output image in degrees. \
    Default: 0.0.
    :u*param crop_edges: When a rotation is applied to any image,\
    the result will contain zeros around the edges, which can be removed by \
    cropping the edges by a specified number of pixels. Default: 0
    :u*param auto_crop: If activated, this feature will automatically crop \
    the image to eliminate any regions without data (because of the rotation).\
    Default: False
    :param use_auto_centre: This parameter automatically sets the centre \
    of rotation to the centre of the image. If set to False, the values \
    from centre_x and centre_y are used. Note: The centre needs to be within \
    the image dimensions. Default: True
    :param centre_x: If not use_auto_centre, this value determines the \
    detector x coordinate for the centre of rotation. The Default: 1279.5
    :param centre_y: If not use_auto_centre, this value determines the \
    detector x coordinate for the centre of rotation. Default: 1079.5
    """

    def __init__(self):
        super(CameraRotCorrection, self).__init__("CameraRotCorrection")

    def pre_process(self):
        pass

    def process_frames(self, data):
        return rotate(data[0].astype(np.float64), self.parameters['angle'], center=self.centre)[tuple(self.new_slice)]

    def post_process(self):
        pass

    def setup(self):
        in_dataset, out_dataset = self.get_datasets()
        in_pData, out_pData = self.get_plugin_datasets()
        in_pData[0].plugin_data_setup('PROJECTION', 'single')
        det_y = in_dataset[0].get_data_dimension_by_axis_label('detector_y')
        det_x = in_dataset[0].get_data_dimension_by_axis_label('detector_x')

        self.shape = list(in_dataset[0].get_shape())
        self.core_dims = in_pData[0].get_core_dimensions()

        self.use_auto_centre = self.parameters['use_auto_centre']
        self.angle = self.parameters['angle']
        self.static_crop = self.parameters['crop_edges']
        self.auto_crop = self.parameters['auto_crop']

        if self.auto_crop and isinstance(self.angle, list):
            plugin = self.__class__.__name__
            msg = "Parameter tuning on the angles in %s plugin is only " \
                  "possible with auto_crop set to False." % plugin
            raise Exception(msg)

        img_dims = self.get_in_datasets()[0].get_shape()
        if self.use_auto_centre:
            self.centre = (img_dims[det_y] // 2 - 0.5 * np.mod(img_dims[det_y], 2),
                           img_dims[det_x] // 2 - 0.5 * np.mod(img_dims[det_x], 2))
        else:
            self.centre = (self.parameters['centre_y'], self.parameters['centre_x'])

        # If the data is cropped then the centre of rotation must be shifted
        # accordingly, e.g if preview is [:, a:b, c:d] then shift is (a, c)
        shift = self.exp.meta_data.get(in_dataset[0].get_name() + '_preview_starts')
        self.centre = (self.centre[0] - shift[det_y], self.centre[1] - shift[det_x])

        self.new_slice = [slice(None)] * 2
        img_dims = self.get_in_datasets()[0].get_shape()

        if self.static_crop > 0 and not self.auto_crop:
            self.new_slice = [slice(self.static_crop, img_dims[det_y] - self.static_crop),
                              slice(self.static_crop, img_dims[det_x] - self.static_crop)]
            for ddim in [det_x, det_y]:
                self.shape[ddim] = self.shape[ddim] - 2 * self.static_crop

        elif self.auto_crop:
            xs = np.array([0, 0, img_dims[det_x], img_dims[det_x]]).astype(np.float32)
            ys = np.array([0, img_dims[det_x], img_dims[det_x], 0]).astype(np.float32)
            r = np.sqrt((xs - self.centre[1]) ** 2 + (ys - self.centre[0]) ** 2)
            theta = np.pi + np.arctan2(xs - self.centre[1], ys - self.centre[0])
            theta_p = theta + self.angle / 180. * np.pi
            x_p = np.sort(self.centre[1] - np.sin(theta_p) * r)
            y_p = np.sort(self.centre[0] - np.cos(theta_p) * r)
            x0 = np.ceil(max(x_p[0], x_p[1], 0)).astype(int)
            x1 = np.floor(min(x_p[2], x_p[3], img_dims[det_x])).astype(int)
            y0 = np.ceil(max(y_p[0], y_p[1], 0)).astype(int)
            y1 = np.floor(min(y_p[2], y_p[3], img_dims[det_y])).astype(int)
            self.new_slice = [slice(y0, y1), slice(x0, x1)]
            self.shape[det_x] = x1 - x0
            self.shape[det_y] = y1 - y0

        out_dataset[0].create_dataset(patterns=in_dataset[0],
                                      axis_labels=in_dataset[0],
                                      shape=tuple(self.shape))
        out_pData[0].plugin_data_setup('PROJECTION', 'single')




# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: image_saver
   :platform: Unix
   :synopsis: A class to save output as images

.. moduleauthor:: Dan Nixon, Nghia Vo <scientificsoftware@diamond.ac.uk>

"""

import skimage.exposure
import skimage.io
import numpy as np
from PIL import Image

from savu.plugins.savers.base_image_saver import BaseImageSaver
from savu.plugins.utils import register_plugin
from savu.plugins.driver.cpu_plugin import CpuPlugin
import savu.core.utils as cu


@register_plugin
class ImageSaver(BaseImageSaver, CpuPlugin):
    """
    A class to save tomography data to image files.  Run the MaxAndMin plugin\
    before this to rescale the data.

    :param pattern: How to slice the data. Default: 'VOLUME_XZ'.
    :u*param format: Image format. Default: 'tif'.
    :u*param num_bit: Bit depth of the tiff format (8, 16 or 32). Default: 16.
    :param max: Global max for tiff scaling. Default: None.
    :param min: Global min for tiff scaling. Default: None.
    :param jpeg_quality: JPEG encoding quality (1 is worst, 100 is best). Default: 75.
    :param prefix: Override the default output jpg file prefix. Default: None.

    :config_warn: Do not use this plugin if the raw data is greater than \
    100 GB.
    """

    def __init__(self, name='ImageSaver'):
        super(ImageSaver, self).__init__(name)

    def pre_process(self):
        super(ImageSaver, self).pre_process()
        self.file_format = self.parameters['format']
        num_bit = self.parameters['num_bit']
        self.pData = self.get_plugin_in_datasets()[0]
        if not (num_bit == 8 or num_bit == 16 or num_bit == 32):
            self.num_bit = 32
            msg = "\n***********************************************\n"\
                "This option %s is not available. Reset to 32 \n"\
                % str(num_bit)
            cu.user_message(msg)
        else:
            self.num_bit = num_bit
        self._data_range = self._get_min_and_max()

    def process_frames(self, data):
        frame = self.pData.get_current_frame_idx()[0]
        filename = '%s%05i.%s' % (self.filename, frame, self.file_format)
        frame = np.nan_to_num(data[0])
        if (self.file_format == "tiff") or (self.file_format == "tif"):
            global_min = self.parameters['min']
            global_max = self.parameters['max']
            if self.num_bit == 32:
                rescaled_image = frame
            else:
                if global_min is None:
                    if self.the_min is not None:
                        global_min = self.the_min
                    else:
                        global_min = np.min(frame)
                if global_max is None:
                    if self.the_max is not None:
                        global_max = self.the_max
                    else:
                        global_max = np.max(frame)
                rescaled_image = np.clip(frame, global_min, global_max)
                rescaled_image = (rescaled_image - global_min) \
                    / (global_max - global_min)
                if self.num_bit == 16:
                    rescaled_image = np.clip(
                        np.uint16(rescaled_image * 65535), 0, 65535)
                else:
                    rescaled_image = np.clip(
                        np.uint8(rescaled_image * 255), 0, 255)
            img = Image.fromarray(rescaled_image)
            img.save(filename)
        else:
            # Rescale image to (0.0, 1.0) range
            rescaled_image = skimage.exposure.rescale_intensity(
                frame, in_range=self._data_range, out_range=(0.0, 1.0))
            # Save image
            skimage.io.imsave(
                filename, rescaled_image, quality=self.parameters['jpeg_quality'])

    def _get_min_and_max(self):
        data = self.get_in_datasets()[0]
        pattern = self.parameters['pattern']
        try:
            self.the_min = np.min(
                data.meta_data.get(['stats', 'min', pattern]))
            self.the_max = np.max(
                data.meta_data.get(['stats', 'max', pattern]))
            self._data_range = (self.the_min, self.the_max)
        except KeyError:
            self._data_range = 'image'
            if (self.file_format == "tiff") or (self.file_format == "tif"):
                self.the_min = None
                self.the_max = None
                msg = "\n***********************************************\n"\
                    "!!!Warning!!!-> No global maximum and global minimum found\n"\
                    "in the metadata. Please run the MaxAndMin plugin before\n" \
                    "ImageSaver or input manually. Otherwise, local minimum\n" \
                    "and local maximum will be used for rescaling. This may\n"\
                    "result the fluctuation of brightness between slices.\n"\
                    "***********************************************\n"
                if (self.num_bit == 8) or (self.num_bit == 16):
                    cu.user_message(msg)
        return self._data_range

    def executive_summary(self):
        if self._data_range == 'image':
            return ["To rescale and normalise the data between global max and "
                    "min values, please run MaxAndMin plugin before "
                    "ImageSaver."]
        return ["Nothing to Report"]



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: band_pass
   :platform: Unix
   :synopsis: A plugin to low pass each frame with a gaussian

.. moduleauthor:: Mark Basham <scientificsoftware@diamond.ac.uk>

"""
import logging

from savu.plugins.filters.base_filter import BaseFilter
from savu.plugins.driver.cpu_plugin import CpuPlugin

from scipy.ndimage.filters import gaussian_filter

from savu.plugins.utils import register_plugin


@register_plugin
class BandPass(BaseFilter, CpuPlugin):
    """
    A plugin to filter each frame with a BandPass T

    :param blur_width: Kernel Size. Default: (0 ,3 ,3).
    :param type: filter type (High|Low). Default: 'High'.
    """

    def __init__(self):
        """
        This is where the class is designed, and we inherit from 2 classes, the
        first is the Filter class, which deals with splitting the job up for
        us, and provides some simple methods which we need to overload. The
        second is the CpuPlugin class which tells the framework that the
        processing that you are doing here runs on 1 cpu.
        """
        logging.debug("initialising BandPass filter")
        logging.debug("Calling super to make sure that all superclases are " +
                      " initialised")
        super(BandPass,
              self).__init__("BandPass")

    def process_frames(self, data):
        """
        The second method we need to implement from the Filter class and the
        part of the code that actually does all the work. the input here 'data'
        will contain the 3D block of data to process, and we need to return the
        data for the single frame in the middle of this. In this case we use
        the scipy median filter with the 'kernmel_size' parameter, and return
        the same size data as you had originally.
        """
        data = data[0]
        logging.debug("Data frame recieved for processing of shape %s",
                      str(data.shape))
        if self.parameters['type'] == 'Low':
            result = gaussian_filter(data, self.parameters['blur_width'])
        elif self.parameters['type'] == 'High':
            lp = gaussian_filter(data, self.parameters['blur_width'])
            result = data - lp
        return result



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: simple_recon
   :platform: Unix
   :synopsis: A simple implementation of a reconstruction routine for testing
       purposes

.. moduleauthor:: Mark Basham <scientificsoftware@diamond.ac.uk>

"""
from savu.plugins.reconstructions.base_recon import BaseRecon
from savu.data.plugin_list import CitationInformation
from savu.plugins.driver.cpu_plugin import CpuPlugin

import numpy as np

from savu.plugins.utils import register_plugin


@register_plugin
class SimpleRecon(BaseRecon, CpuPlugin):
    """
    A Plugin to apply a simple reconstruction with no dependancies
    """

    def __init__(self):
        super(SimpleRecon, self).__init__("SimpleRecon")

    def _filter(self, sinogram):
        ff = np.arange(sinogram.shape[0])
        ff -= sinogram.shape[0] // 2
        ff = np.abs(ff)
        fs = np.fft.fft(sinogram)
        ffs = fs*ff
        return np.fft.ifft(ffs).real

    def _back_project(self, mapping, sino_element, centre):
        mapping_array = mapping+centre
        return sino_element[mapping_array.astype('int')]

    def _mapping_array(self, shape, center, theta):
        x, y = np.meshgrid(np.arange(-center[0], shape[0] - center[0]),
                           np.arange(-center[1], shape[1] - center[1]))
        return x*np.cos(theta) - y*np.sin(theta)

    def process_frames(self, data):
        sino = data[0]
        centre_of_rotations, angles, vol_shape, init = self.get_frame_params()
        sinogram = sino[:, np.newaxis, :]
        try:
            centre = self.kwargs['centre']
        except Exception:
            centre = (vol_shape[0] // 2, vol_shape[1] // 2)

        results = []
        for j in range(sinogram.shape[1]):
            result = np.zeros(vol_shape, dtype=np.float32)
            for i in range(sinogram.shape[0]):
                theta = i*(np.pi/sinogram.shape[0])
                mapping_array = self._mapping_array(vol_shape, centre, theta)
                filt = np.zeros(sinogram.shape[2]*3, dtype=np.float32)
                filt[sinogram.shape[2]:sinogram.shape[2]*2] = \
                    self._filter(np.log(np.nan_to_num(sinogram[i, j, :])+1))
                result += \
                    self._back_project(mapping_array, filt,
                                       (centre_of_rotations +
                                           sinogram.shape[2]))
            results.append(result[:, np.newaxis, :])
        result = np.hstack(results)
        return result

    def get_max_frames(self):
        return 'single'

    def get_citation_information(self):
        cite_info = CitationInformation()
        cite_info.description = \
            ("The Tomographic reconstruction performed in this processing " +
             "chain is derived from this work.")
        cite_info.bibtex = \
            ("@book{avinash2001principles,\n" +
             "  title={Principles of computerized tomographic imaging},\n" +
             "  author={Avinash C.. Kak and Slaney, Malcolm},\n" +
             "  year={2001},\n" +
             "  publisher={Society for Industrial and Applied Mathematics}\n" +
             "}")
        cite_info.endnote = \
            ("%0 Book\n" +
             "%T Principles of computerized tomographic imaging\n" +
             "%A Avinash C.. Kak\n" +
             "%A Slaney, Malcolm\n" +
             "%@ 089871494X\n" +
             "%D 2001\n" +
             "%I Society for Industrial and Applied Mathematics")
        cite_info.doi = "https://doi.org/10.1137/1.9780898719277"
        return cite_info



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: pyfai_azimuthal_integrator_separate
   :platform: Unix
   :synopsis: A plugin to integrate azimuthally "symmetric" signals \
       i.e. SAXS, WAXS or XRD.Requires a calibration file

.. moduleauthor:: Aaron D. Parsons <scientificsoftware@diamond.ac.uk>
"""

import logging
from savu.plugins.azimuthal_integrators.base_azimuthal_integrator import \
    BaseAzimuthalIntegrator
from savu.plugins.utils import register_plugin


@register_plugin
class PyfaiAzimuthalIntegratorSeparate(BaseAzimuthalIntegrator):
    """
    1D azimuthal integrator by pyFAI
    :param use_mask: Should we mask. Default: False.
    :param out_datasets: A. Default: ["powder","spots"]
    :param percentile: Percentile to threshold. Default:50.
    :param num_bins: number of bins. Default: 1005.
    :param num_bins_azim: number of azimuthal bins. Default:20.

    """

    def __init__(self):
        logging.debug("Starting 1D azimuthal integrationr")
        super(PyfaiAzimuthalIntegratorSeparate,
              self).__init__("PyfaiAzimuthalIntegratorSeparate")

    def process_frames(self, data):
        logging.debug("Running azimuthal integration")
        mData = self.params[2]
        mask = self.params[0]
        ai = self.params[3]
        num_bins_azim = self.parameters['num_bins_azim']
        num_bins_rad = self.parameters['num_bins']
        percentile = self.parameters['percentile']
        spots, powder = ai.separate(data[0], npt_rad=num_bins_rad, npt_azim=num_bins_azim, percentile=percentile)
        axis, spectra = ai.integrate1d(data=powder,npt=num_bins_rad, unit='q_A^-1')
        return [spectra, spots]

    def setup(self):
        in_dataset, out_datasets = self.get_datasets()
        in_pData, out_pData = self.get_plugin_datasets()
        shape = in_dataset[0].get_shape()
        # it will always be in Q for this plugin
        # Doesnt this get rid of the other two axes?
        #axis_labels = {in_dataset[0]: '-1.Q.nm^-1'}
        # I just want diffraction data
        in_pData[0].plugin_data_setup('DIFFRACTION', self.get_max_frames())
        spots = out_datasets[1]
        spectra = out_datasets[0]
        num_bins = self.get_parameters('num_bins')
        # what does this do?
        #remove an axis from all patterns

        # copy all patterns, removing dimension -1 from the core and slice
        # directions, and returning only those that are not empty
        patterns = ['SINOGRAM.-1', 'PROJECTION.-1']
        # stating only 'dimension' will remove the axis label, stating
        # 'dimension.name.unit' name and unit will add or replace it
        axis_labels = ['-1', '-2.name.unit']
        spots.create_dataset(in_dataset[0])
        spectra.create_dataset(patterns={in_dataset[0]: patterns},
                               axis_labels={in_dataset[0]: axis_labels},
                               shape=shape[:-2]+(num_bins,))

        spectrum = \
            {'core_dims': (-1,), 'slice_dims': tuple(range(len(shape)-2))}
        spectra.add_pattern("SPECTRUM", **spectrum)

        out_pData[0].plugin_data_setup('SPECTRUM', self.get_max_frames())
        out_pData[1].plugin_data_setup('DIFFRACTION', self.get_max_frames())

    def nOutput_datasets(self):
        return 2



#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Created on Tue Oct 29 15:42:30 2019

@author: qmm55171
"""

import os
import sys
import argparse

from savu.data.meta_data import MetaData
from scripts.config_generator.content import Content
from savu.plugins.loaders.yaml_converter import YamlConverter


def arg_parser():
    """ Arg parser for command line arguments.
    """
    parser = argparse.ArgumentParser(prog='update_template')
    parser.add_argument('nxsfile', help='The nexus file')
    key_help = 'Full stop separated key for dictionary entry'
    parser.add_argument("-k", "--key", help=key_help, default=None)
    return parser.parse_args()


def get_parameter_value(nxsfile, plugin_name, plugin_param):
    content = Content()
    content.fopen(nxsfile)
    plist = content.plugin_list.plugin_list

    found_params = []
    for plugin in plist:
        if plugin['name'] == plugin_name:
            if plugin_param in list(plugin['data'].keys()):
                found_params.append(plugin['data'][plugin_param])

    return found_params


def check_yaml_path(paths):
    paths = [os.path.abspath(p) for p in paths]
    for p in paths:
        if not os.path.isfile(p):
            raise Exception("Template file not found at %s" % p)
    return paths


def get_data_dict(paths):
    all_data_dict = {}
    yu = YamlConverter()
    for p in paths:
        yu.parameters['yaml_file'] = p
        all_data_dict.update(yu.setup(template=True))
    return all_data_dict


def get_string_result(key, data_dict, mData, res=[]):
    if '*' in key:
        for data in list(data_dict.keys()):
            res = get_string_result(
                    key.replace('*', data), data_dict, mData, res)
    else:
        res.append(mData.get(key.split('.')))
    return res


if __name__ == '__main__':
    arg = arg_parser()
    # find all Hdf5TemplateLoaders
    plugin = 'Hdf5TemplateLoader'
    param = 'yaml_file'
    paths = check_yaml_path(get_parameter_value(arg.nxsfile, plugin, param))
    data_dict = get_data_dict(paths)
    res = get_string_result(arg.key, data_dict, MetaData(data_dict))
    print(','.join(res))

    


# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: Point-spread-function correction
   :platform: Unix
   :synopsis: A plugin for MTF (modulation transfer function) deconvolution\
    or PSF (point spread function) correction in the Fourier domain.
.. moduleauthor:: Nghia Vo <scientificsoftware@diamond.ac.uk>

"""
import os
import logging
import numpy as np
from PIL import Image
import pyfftw.interfaces.scipy_fftpack as fft
from savu.plugins.plugin import Plugin
from savu.plugins.driver.cpu_plugin import CpuPlugin
from savu.plugins.utils import register_plugin
from savu.data.plugin_list import CitationInformation
import savu.core.utils as cu



@register_plugin
class MtfDeconvolution(Plugin, CpuPlugin):
    """
    Method to correct the point-spread-function effect. \
    Working on raw projections and flats.
    :u*param file_path: Path to file containing a 2D array of a MTF function. \
        File formats: 'npy', or 'tif'. Default: None.
    :param pad_width: Pad the image before the deconvolution. Default: 128.

    """

    def __init__(self):
        super(MtfDeconvolution, self).__init__("MtfDeconvolution")

    def load_image(self, file_path):
        """
        Load data from an image.
        """
        mat = None
        try:
            mat = np.asarray(Image.open(file_path), dtype = np.float32)
        except IOError:
            raise ValueError("No such file or directory {}".format(file_path))
        if len(mat.shape) > 2:
            axis_m = np.argmin(mat.shape)
            mat = np.mean(mat, axis=axis_m)
        return mat

    def check_file_path(self, file_path):
        file_ext = ".tif"
        if file_path is None:
            msg = "!!! Please provide a file path to the MTF !!!"
            logging.warning(msg)
            cu.user_message(msg)
            raise ValueError(msg)
        else:
            if not os.path.isfile(file_path):
                msg = "!!! No such file: %s !!!"\
                        " Please check the file path" %str(file_path)
                logging.warning(msg)
                cu.user_message(msg)
                raise ValueError(msg)
            else:
                _, file_ext = os.path.splitext(file_path)
        return file_ext

    def psf_correction(self, mat, win, pad_width):
        (nrow, ncol) = mat.shape
        mat_pad = np.pad(mat, pad_width, mode = "reflect")
        win_pad = np.pad(win, pad_width, mode = "constant", constant_values=1.0)
        mat_dec = fft.ifft2(fft.fft2(mat_pad) / fft.ifftshift(win_pad))
        return np.abs(mat_dec)[pad_width:pad_width+nrow,pad_width:pad_width+ncol]

    def setup(self):
        in_dataset, out_dataset = self.get_datasets()
        out_dataset[0].create_dataset(in_dataset[0], raw=True)
        in_pData, out_pData = self.get_plugin_datasets()
        in_pData[0].plugin_data_setup('PROJECTION','single')
        out_pData[0].plugin_data_setup('PROJECTION','single')

    def pre_process(self):
        inData = self.get_in_datasets()[0]
        dark = inData.data.dark()
        flat = inData.data.flat()
        self.data_size = inData.get_shape()
        (self.depth, self.height, self.width) = flat.shape
        file_path = self.parameters["file_path"]
        file_ext = self.check_file_path(file_path)
        if file_ext==".npy":
            try:
                self.mtf_array = 1.0*np.load(file_path)
            except IOError:
                msg = "\n*****************************************\n"\
                    "!!! ERROR !!! -> Can't open this file: %s \n"\
                    "*****************************************\n\
                    " % str(file_path)
                logging.warning(msg)
                cu.user_message(msg)
                raise ValueError(msg)
        else:
            try:
                self.mtf_array = 1.0*np.float32(self.load_image(file_path))
            except IOError:
                msg = "\n*****************************************\n"\
                    "!!! ERROR !!! -> Can't open this file: %s \n"\
                    "*****************************************\n\
                    " % str(file_path)
                logging.warning(msg)
                cu.user_message(msg)
                raise ValueError(msg)

        self.mtf_array[self.mtf_array<=0.0] = 1.0
        self.mtf_array = self.mtf_array/np.max(self.mtf_array)
        (height_mtf, width_mtf) = self.mtf_array.shape
        if (self.height != height_mtf) or (self.width != width_mtf):
            msg = "\n*****************************************\n"\
            "!!! ERROR !!!-> Projection shape: ({0},{1}) is not the same as "\
            "the mtf shape: ({2},{3})".format(
                self.height, self.width, height_mtf, width_mtf)
            logging.warning(msg)
            cu.user_message(msg)
            raise ValueError(msg)

        self.pad_width = np.clip(int(self.parameters["pad_width"]), 0, None)
        if flat.size:
            flat_updated = np.ones_like(flat, dtype=np.float32)
            for i in np.arange(self.depth):
                flat_updated[i] = self.psf_correction(
                    flat[i], self.mtf_array, self.pad_width)
            inData.data.update_flat(flat_updated)

    def process_frames(self, data):
        return self.psf_correction(data[0], self.mtf_array, self.pad_width)

    def get_citation_information(self):
        cite_info = CitationInformation()
        cite_info.description = \
            ("The PSF correction used in this plugin is taken\
             from this work.")
        cite_info.bibtex = ("@inproceedings{10.1117/12.2530324,\n"\
            "author = {Nghia T. Vo and Robert C. Atwood "\
            "and Michael Drakopoulos},\n"\
            "title = {{Preprocessing techniques for removing artifacts in "\
            "synchrotron-based tomographic images}},\n"\
            "volume = {11113},\n"\
            "booktitle = {Developments in X-Ray Tomography XII},\n"\
            "editor = {Bert Muller and Ge Wang},\n"\
            "organization = {International Society for Optics and Photonics},\n"\
            "publisher = {SPIE},\n"\
            "pages = {309 -- 328},\n"\
            "year = {2019},\n"\
            "doi = {10.1117/12.2530324},\n"\
            "URL = {https://doi.org/10.1117/12.2530324}\n"\
            "}")
        cite_info.doi = "doi: DOI: 10.1117/12.2530324"
        return cite_info



# Copyright 2019 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: Gaussian mixture models for classification-segmentation routine
   :platform: Unix
   :synopsis: Wrapper around scikit GMM function

.. moduleauthor:: Daniil Kazantsev <scientificsoftware@diamond.ac.uk>
"""

from savu.plugins.plugin import Plugin
from savu.plugins.driver.multi_threaded_plugin import MultiThreadedPlugin
from savu.plugins.utils import register_plugin

import numpy as np
from sklearn.mixture import GaussianMixture

@register_plugin
class GmmSegment3d(Plugin, MultiThreadedPlugin):
    """
    A Plugin to segment data using Gaussian Mixtures from scikit

    :param classes: The number of classes for GMM. Default: 4.
    """

    def __init__(self):
        super(GmmSegment3d, self).__init__("GmmSegment3d")

    def setup(self):

        in_dataset, out_dataset = self.get_datasets()
        out_dataset[0].create_dataset(in_dataset[0], dtype=np.uint8)
        in_pData, out_pData = self.get_plugin_datasets()

        getall = ["VOLUME_XZ", "voxel_y"]
        in_pData[0].plugin_data_setup('VOLUME_3D', 'single', getall=getall)
        out_pData[0].plugin_data_setup('VOLUME_3D', 'single', getall=getall)

    def pre_process(self):
        # extract given parameters
        self.classes = self.parameters['classes']

    def process_frames(self, data):
        # Do GMM classification/segmentation first
        dimensdata = data[0].ndim
        if (dimensdata == 2):
            (Nsize1, Nsize2) = np.shape(data[0])
            Nsize3 = 1
        if (dimensdata == 3):
            (Nsize1, Nsize2, Nsize3) = np.shape(data[0])

        inputdata = data[0].reshape((Nsize1*Nsize2*Nsize3), 1)/np.max(data[0])

        #run classification and segmentation
        classif = GaussianMixture(n_components=self.classes, covariance_type="tied")
        classif.fit(inputdata)
        cluster = classif.predict(inputdata)
        segm = classif.means_[cluster]
        if (dimensdata == 2):
            segm = segm.reshape(Nsize1, Nsize3, Nsize2)
        else:
            segm = segm.reshape(Nsize1, Nsize2, Nsize3)
        maskGMM = segm.astype(np.float64) / np.max(segm)
        maskGMM = 255 * maskGMM # Now scale by 255
        maskGMM = maskGMM.astype(np.uint8) # obtain the GMM mask
        return [maskGMM]

    def nInput_datasets(self):
        return 1
    def nOutput_datasets(self):
        return 1



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: temp_loader
   :platform: Unix
   :synopsis: A class for loading standard tomography data in a variety of
    formats.

.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>

"""

import os
import tempfile
import warnings

import h5py
import numpy as np

from savu.data.data_structures.data_types.data_plus_darks_and_flats \
    import NoImageKey
from savu.data.data_structures.data_types.image_data import ImageData
from savu.plugins.loaders.base_loader import BaseLoader
from savu.plugins.utils import register_plugin


@register_plugin
class ImageLoader(BaseLoader):
    """
    Load any FabIO compatible formats (e.g. tiffs)

    :param dataset_name: The name assigned to the dataset. Default: 'tomo'.
    :param angles: A python statement to be evaluated \
    (e.g np.linspace(0, 180, nAngles)) or a file. Default: None.
    :param frame_dim: Which dimension requires stitching? Default: 0.
    :param data_prefix: A file prefix for the data file. Default: None.
    :param dark_prefix: A file prefix for the dark field files, including the\
    folder path if different from the data. Default: None.
    :param flat_prefix: A file prefix for the flat field files, including the\
    folder path if different from the data. Default: None.
    """

    def __init__(self, name='ImageLoader'):
        super(ImageLoader, self).__init__(name)

    def setup(self):
        exp = self.exp
        data_obj = exp.create_data_object('in_data',
                                          self.parameters['dataset_name'])

        rot = 0
        detY = 1
        detX = 2
        data_obj.set_axis_labels('rotation_angle.degrees',
                                 'detector_y.pixel',
                                 'detector_x.pixel')

        data_obj.add_pattern('PROJECTION', core_dims=(detX, detY),
                             slice_dims=(rot,))
        data_obj.add_pattern('SINOGRAM', core_dims=(detX, rot),
                             slice_dims=(detY,))

        path = os.path.abspath(exp.meta_data.get("data_file"))
        data_obj.data = self._get_data_type(data_obj, path)

        self.set_rotation_angles(data_obj)

        # dummy file
        filename = path.split(os.sep)[-1] + '.h5'
        data_obj.backing_file = \
            h5py.File(os.path.join(tempfile.mkdtemp(), filename), 'a')

        data_obj.set_shape(data_obj.data.get_shape())
        self.set_data_reduction_params(data_obj)
        self._set_darks_and_flats(data_obj, path)

    def _set_darks_and_flats(self, dObj, path):
        if not self.parameters['flat_prefix']:
            return

        dObj.data = NoImageKey(dObj, None, 0)
        fdim = self.parameters['frame_dim']

        # read dark and flat images
        fpath, ffix = self._get_path(self.parameters['flat_prefix'], path)
        flat = ImageData(fpath, dObj, [fdim], None, ffix)

        if self.parameters['dark_prefix']:
            dpath, dfix = self._get_path(self.parameters['dark_prefix'], path)
            dark = ImageData(dpath, dObj, [fdim], None, dfix)
        else:
            shape = dObj.get_shape()
            dark = np.zeros([1] + [shape[i] for i in [1, 2]], dtype=flat.dtype)

        dObj.data._set_dark_path(dark)
        dObj.data._set_flat_path(flat)
        dObj.data._set_dark_and_flat()

    def _get_path(self, param, data):
        ppath, pfix = os.path.split(param)
        ppath = \
            os.path.join(data, ppath) if not os.path.isabs(ppath) else ppath
        return ppath, pfix

    def _get_data_type(self, obj, path):
        prefix = self.parameters['data_prefix']
        return ImageData(path, obj, [self.parameters['frame_dim']], None, prefix)

    def set_rotation_angles(self, data_obj):
        angles = self.parameters['angles']

        if angles is None:
            angles = np.linspace(0, 180, data_obj.data.get_shape()[0])
        else:
            try:
                exec("angles = " + angles)
            except Exception as e:
                warnings.warn("Could not execute statement: {}".format(e))
                try:
                    angles = np.loadtxt(angles)
                except Exception as e:
                    raise Exception('Cannot set angles in loader. Error: {}'.format(e))

        n_angles = len(angles)
        data_angles = data_obj.data.get_shape()[self.parameters['frame_dim']]
        if data_angles != n_angles:
            raise Exception("The number of angles %s does not match the data "
                            "dimension length %s", n_angles, data_angles)
        data_obj.meta_data.set("rotation_angle", angles)



# -*- coding: utf-8 -*-
# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
.. module:: plugins_test
   :platform: Unix
   :synopsis: unittest test classes for plugins

.. moduleauthor:: Mark Basham <scientificsoftware@diamond.ac.uk>

"""

import unittest
import savu.test.test_utils as tu
from savu.test.travis.framework_tests.plugin_runner_test import \
    run_protected_plugin_runner


class MultipleParameterProcessListTests(unittest.TestCase):

    def test_multi_params_tomo(self):
        data_file = tu.get_test_data_path('24737.nxs')
        process_file = tu.get_test_process_path(
            'basic_tomo_process_preview_params_test.nxs')
        run_protected_plugin_runner(tu.set_options(data_file,
                                                   process_file=process_file))

#    def test_multi_params_i12tomo(self):
#        process = 'i12_tomo_pipeline_test.nxs'
#        options = {
#            "transport": "hdf5",
#            "process_names": "CPU0",
#            "data_file": tu.get_test_data_path(
#                'i12_test_data/i12_test_data.nxs'),
#            "process_file": tu.get_test_process_path(process),
#            "out_path": tempfile.mkdtemp()
#            }
#        run_protected_plugin_runner(options)

if __name__ == "__main__":
    unittest.main()



# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
.. module:: plugin_template6
   :platform: Unix
   :synopsis: A template to create a plugin that changes the shape of the data.

.. moduleauthor:: Developer Name <email@address.ac.uk>

"""
import numpy as np

from savu.plugins.plugin import Plugin
from savu.plugins.driver.cpu_plugin import CpuPlugin
from savu.plugins.utils import register_plugin


@register_plugin
class PluginTemplate6(Plugin, CpuPlugin):
    """
    A template to create a plugin that changes the shape of the data, \
    e.g. downsample_filter.

    :param example: Example of a plugin parameter. Default: None.
    :u*param bin_size: Bin Size for the downsample. Default: 2.
    :u*param pattern: The pattern the plugin should be applied \
    to. Default: "PROJECTION".

    """

    def __init__(self):
        super(PluginTemplate6, self).__init__('PluginTemplate6')

    def nInput_datasets(self):
        return 1

    def nOutput_datasets(self):
        return 1

    def setup(self):
        # get all in and out datasets required by the plugin
        in_dataset, self.out_dataset = self.get_datasets()

        # set in_plugin_dataset first so pattern information is available to
        # calculate the new shape
        self.in_pData, self.out_pData = self.get_plugin_datasets()
        pattern = self.parameters['pattern']
        self.in_pData[0].plugin_data_setup(pattern, 'single')

        # calculate the output shape (based on the input shape)
        self.out_shape = \
            self.new_shape(in_dataset[0].get_shape(), in_dataset[0])

        #=================== populate output datasets =========================
        # the output data shape retains the same patterns and axis labels but
        # requires a different shape.
        self.out_dataset[0].create_dataset(patterns=in_dataset[0],
                                           axis_labels=in_dataset[0],
                                           shape=self.out_shape)

        #================== populate output plugin datasets ===================
        self.out_pData[0].plugin_data_setup(pattern, 'single')

    def new_shape(self, full_shape, data):
        # example of a function to calculate a new output data shape based on
        # the input data shape
        core_dirs = data.get_core_dimensions()
        new_shape = list(full_shape)
        for dim in core_dirs:
            new_shape[dim] = full_shape[dim] // self.parameters['bin_size']
        return tuple(new_shape)

    def pre_process(self):
        # Example of calculating a new slice list to reduce the data
        in_data_shape = self.in_pData[0].get_shape()
        bin_size = self.parameters['bin_size']
        new_sl = [slice(0, i, bin_size) for i in in_data_shape]

        # update all axis label values based on the new slice list
        self.out_dataset[0].amend_axis_label_values(
            self.out_pData[0]._get_data_slice_list(new_sl))

    def process_frames(self, data):
        # replace this with your function
        return np.zeros(self.get_plugin_out_datasets()[0].get_shape())

    def post_process(self):
        pass



# -*- coding: utf-8 -*-
# Copyright 2014 Diamond Light Source Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
.. module:: i12_data_mapping_test
   :platform: Unix
   :synopsis: unittest test classes for plugins

.. moduleauthor:: Nicola Wadeson <scientificsoftware@diamond.ac.uk>

"""

import unittest

import savu.test.test_utils as tu
from savu.test.travis.framework_tests.plugin_runner_test import \
    run_protected_plugin_runner_no_process_list


class Mapping3Dto4D(unittest.TestCase):

    def get_loader_dict(self):
        return {'data_path': '/1-TempPlugin-tomo/data',
                'angles': 'np.linspace(0, 180, 181)',
                '3d_to_4d': True}

    def test_3dto4d(self):
        options = tu.set_experiment('tomo_3dto4d')
        #plugin = 'savu.plugins.corrections.i12_dark_flat_field_correction'
        plugin = 'savu.plugins.filters.no_process_plugin'
        data_dict = {'in_datasets': ['tomo'], 'out_datasets': ['test']}
        all_dicts = [self.get_loader_dict(), data_dict, {}]
        exp = run_protected_plugin_runner_no_process_list(options, plugin,
                                                          data=all_dicts)
        self.assertEqual(exp.index['in_data']['test'].get_shape(),
                         (181, 10, 192, 4))

    def test_3dto4d_previewing1(self):
        options = tu.set_experiment('tomo_3dto4d')
        plugin = 'savu.plugins.filters.no_process_plugin'
        selection = \
            ['mid:mid+1:1:5', '0:end:1:1', '0:end:1:1', 'mid:mid+1:1:1']
        loader_dict = self.get_loader_dict()
        loader_dict['preview'] = selection
        data_dict = {'in_datasets': ['tomo'], 'out_datasets': ['test']}
        all_dicts = [loader_dict, data_dict, {}]
        exp = run_protected_plugin_runner_no_process_list(options, plugin,
                                                          data=all_dicts)
        self.assertEqual(exp.index['in_data']['test'].get_shape(),
                         (5, 10, 192, 1))
#
#    def test_i12tomo3(self):
#        options = tu.set_experiment('i12tomo')
#        plugin = 'savu.plugins.corrections.i12_dark_flat_field_correction'
#        selection = ['midmap:4*endmap:endmap:5', '0:end:10:1', '0:end:10:1']
#        loader_dict = self.get_loader_dict()
#        loader_dict['preview'] = selection
#        data_dict = {'in_datasets': ['tomo'], 'out_datasets': ['test']}
#        all_dicts = [loader_dict, data_dict, {}]
#        exp = run_protected_plugin_runner_no_process_list(options, plugin,
#                                                          data=all_dicts)
#        self.assertEqual(exp.index['in_data']['test'].get_shape(),
#                         (5, 1, 20, 4))
#
#    def test_i12tomo4(self):
#        options = tu.set_experiment('i12tomo')
#        plugin = 'savu.plugins.corrections.i12_dark_flat_field_correction'
#        selection = ['midmap:4*endmap:endmap:1', '0:end:10:1', '0:end:10:1']
#        loader_dict = self.get_loader_dict()
#        loader_dict['preview'] = selection
#        data_dict = {'in_datasets': ['tomo'], 'out_datasets': ['test']}
#        all_dicts = [loader_dict, data_dict, {}]
#        exp = run_protected_plugin_runner_no_process_list(options, plugin,
#                                                          data=all_dicts)
#        self.assertEqual(exp.index['in_data']['test'].get_shape(),
#                         (1, 1, 20, 4))
#
#    def test_i12tomo5(self):
#        options = tu.set_experiment('i12tomo')
#        plugin = 'savu.plugins.corrections.i12_dark_flat_field_correction'
#        selection = ['0:20:1:1', '0:-1:5:1', '0:-1:5:1']
#        loader_dict = self.get_loader_dict()
#        loader_dict['preview'] = selection
#        data_dict = {'in_datasets': ['tomo'], 'out_datasets': ['test']}
#        all_dicts = [loader_dict, data_dict, {}]
#        exp = run_protected_plugin_runner_no_process_list(options, plugin,
#                                                          data=all_dicts)
#        self.assertEqual(exp.index['in_data']['test'].get_shape(),
#                         (20, 2, 39, 1))
#
#    def test_i12tomo6(self):
#        options = tu.set_experiment('i12tomo')
#        plugin = 'savu.plugins.corrections.i12_dark_flat_field_correction'
#        selection = ['midmap:end:endmap:5', '0:end:10:1', '0:end:10:1']
#        loader_dict = self.get_loader_dict()
#        loader_dict['preview'] = selection
#        data_dict = {'in_datasets': ['tomo'], 'out_datasets': ['test']}
#        all_dicts = [loader_dict, data_dict, {}]
#        exp = run_protected_plugin_runner_no_process_list(options, plugin,
#                                                          data=all_dicts)
#        self.assertEqual(exp.index['in_data']['test'].get_shape(),
#                         (5, 1, 20, 4))
#
#    def test_i12tomo7(self):
#        options = tu.set_experiment('i12tomo')
#        plugin = 'savu.plugins.corrections.i12_dark_flat_field_correction'
#        selection = ['mid-2:mid+2:1:1', '0:end:10:1', '0:end:10:1']
#        loader_dict = self.get_loader_dict()
#        loader_dict['preview'] = selection
#        data_dict = {'in_datasets': ['tomo'], 'out_datasets': ['test']}
#        all_dicts = [loader_dict, data_dict, {}]
#        exp = run_protected_plugin_runner_no_process_list(options, plugin,
#                                                          data=all_dicts)
#        self.assertEqual(exp.index['in_data']['test'].get_shape(),
#                         (2, 1, 20, 2))
#
#    def test_i12tomo8(self):
#        options = tu.set_experiment('i12tomo')
#        plugin = 'savu.plugins.corrections.i12_dark_flat_field_correction'
#        selection = ['0:endmap:20:1', '0:end:1:1', '0:end:1:1']
#        loader_dict = self.get_loader_dict()
#        loader_dict['preview'] = selection
#        data_dict = {'in_datasets': ['tomo'], 'out_datasets': ['test']}
#        all_dicts = [loader_dict, data_dict, {}]
#        exp = run_protected_plugin_runner_no_process_list(options, plugin,
#                                                          data=all_dicts)
#        self.assertEqual(exp.index['in_data']['test'].get_shape(),
#                         (10, 10, 192, 1))
#
#    def test_i12tomo9(self):
#        options = tu.set_experiment('i12tomo')
#        plugin = 'savu.plugins.corrections.i12_dark_flat_field_correction'
#        selection = ['0:end:1:1', '0:end:30:1', '0:end:300:1']
#        loader_dict = self.get_loader_dict()
#        loader_dict['preview'] = selection
#        data_dict = {'in_datasets': ['tomo'], 'out_datasets': ['test']}
#        all_dicts = [loader_dict, data_dict, {}]
#        exp = run_protected_plugin_runner_no_process_list(options, plugin,
#                                                          data=all_dicts)
#        self.assertEqual(exp.index['in_data']['test'].get_shape(),
#                         (181, 1, 1, 4))


if __name__ == "__main__":
    unittest.main()
